# -*- coding: utf-8 -*-
"""Tests for the benchmark module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest

from msmhelper import benchmark
from msmhelper.statetraj import StateTraj


@pytest.mark.parametrize('trajs, lagtimes, tmax', [
    ([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1], [1, 2], 5),
])
def test_chapman_kolmogorov_test(trajs, lagtimes, tmax):
    """Test Chapman Kolmogorov test."""
    # check float as lag times
    with pytest.raises(TypeError):
        _ = benchmark.ck_test(trajs, lagtimes=[1.2], tmax=tmax)

    # check negative lag times
    with pytest.raises(TypeError):
        _ = benchmark.ck_test(trajs, lagtimes=[-1], tmax=tmax)

    # check 2d lag times
    with pytest.raises(TypeError):
        _ = benchmark.ck_test(trajs, lagtimes=[lagtimes], tmax=tmax)

    # check maximal time negative
    with pytest.raises(TypeError):
        _ = benchmark.ck_test(trajs, lagtimes=lagtimes, tmax=-1)

    # check maximal time float
    with pytest.raises(TypeError):
        _ = benchmark.ck_test(trajs, lagtimes=lagtimes, tmax=5.7)

    # check if all keys exists
    cktest = benchmark.ck_test(trajs, lagtimes=lagtimes, tmax=tmax)
    assert 'md' in cktest
    for lagtime in lagtimes:
        assert lagtime in cktest


@pytest.mark.parametrize('trajs, lagtime, tmax, result', [(
    StateTraj([1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1]), 1, 2,
    {
        'ck': np.array([[0.8, 0.68], [0.8, 0.68]]),
        'time': np.array([1, 2]),
        'is_ergodic': True,
        'is_fuzzy_ergodic': True,
    },
)])
def test__chapman_kolmogorov_test(trajs, lagtime, tmax, result):
    """Test Chapman Kolmogorov test."""
    cktest = benchmark._chapman_kolmogorov_test(trajs, lagtime, tmax)

    np.testing.assert_array_equal(cktest.keys(), result.keys())
    for key in cktest.keys():
        np.testing.assert_array_almost_equal(cktest[key], result[key])


@pytest.mark.parametrize('trajs, lagtime, tmax', [
    ([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1], 2, 5),
])
def test_buchete_hummer_test(trajs, lagtime, tmax):
    """Test Buchete Hummer test."""
    # check float as lag times
    with pytest.raises(TypeError):
        _ = benchmark.bh_test(trajs, lagtime=[1.2], tmax=tmax)

    # check negative lag times
    with pytest.raises(TypeError):
        _ = benchmark.bh_test(trajs, lagtime=-1, tmax=tmax)

    # check maximal time negative
    with pytest.raises(TypeError):
        _ = benchmark.bh_test(trajs, lagtime=lagtime, tmax=-1)

    # check maximal time float
    with pytest.raises(TypeError):
        _ = benchmark.bh_test(trajs, lagtime=lagtime, tmax=5.7)

    # check if all keys exists
    bhtest = benchmark.bh_test(trajs, lagtime=lagtime, tmax=tmax)
    assert 'md' in bhtest
    assert lagtime in bhtest



# -*- coding: utf-8 -*-
"""Create Markov State Model.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

Authors: Daniel Nagel
         Georg Diez

"""
import decorit
import numba
import numpy as np

from msmhelper import linalg, tests
from msmhelper.statetraj import LumpedStateTraj, StateTraj


def estimate_markov_model(trajs, lagtime):
    """Estimates Markov State Model.

    This method estimates the MSM based on the transition count matrix.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    lagtime : int
        Lag time for estimating the markov model given in [frames].

    Returns
    -------
    T : ndarray
        Transition rate matrix.

    permutation : ndarray
        Array with corresponding states.

    """
    trajs = StateTraj(trajs)

    if isinstance(trajs, LumpedStateTraj):
        return _estimate_markov_model(
            trajs.trajs,
            lagtime,
            trajs.nmicrostates,
            trajs.microstates,
        )
    return _estimate_markov_model(
        trajs.trajs,
        lagtime,
        trajs.nstates,
        trajs.states,
    )


def _estimate_markov_model(trajs, lagtime, nstates, perm=None):
    """Estimate MSM based on the transition count matrix."""
    # convert trajs to numba list # noqa: SC100
    if not numba.config.DISABLE_JIT:  # pragma: no cover
        trajs = numba.typed.List(trajs)

    if perm is None:
        perm = np.arange(nstates)

    Tcount = _generate_transition_count_matrix(trajs, lagtime, nstates)
    return row_normalize_matrix(Tcount), perm


@numba.njit
def _generate_transition_count_matrix(trajs, lagtime, nstates):
    """Generate a simple transition count matrix from multiple trajectories."""
    # initialize matrix
    T_count = np.zeros((nstates, nstates), dtype=np.int64)

    for traj in trajs:
        for stateFrom, stateTo in zip(traj[:-lagtime], traj[lagtime:]):
            T_count[stateFrom, stateTo] += 1

    return T_count


@numba.njit
def row_normalize_matrix(mat):
    """Row normalize the given 2d matrix.

    Parameters
    ----------
    mat : ndarray
        Matrix to be row normalized.

    Returns
    -------
    mat : ndarray
        Normalized matrix.

    """
    row_sum = np.sum(mat, axis=1)
    if not row_sum.all():
        row_sum[row_sum == 0] = 1

    # due to missing np.newaxis row_sum[:, np.newaxis] becomes # noqa: SC100
    return mat / row_sum.reshape(mat.shape[0], 1)


def _implied_timescales(tmat, lagtime):
    """
    Calculate implied timescales.

    .. todo::
        - Clearify usage. Better passing trajs to calculate matrix?

    Parameters
    ----------
    tmat : ndarray
        Quadratic transition matrix.

    lagtime: int
        Lagtime for estimating the markov model given in [frames].

    Returns
    -------
    timescales: ndarray
        Implied timescale given in frames.

    """
    tmat = np.asarray(tmat)

    eigenvalues = linalg.left_eigenvalues(tmat)
    # for negative eigenvalues no timescale is defined
    eigenvalues[eigenvalues < 0] = np.nan
    return np.ma.divide(- lagtime, np.log(eigenvalues[1:]))


def implied_timescales(trajs, lagtimes, reversible=False):
    """Calculate the implied timescales.

    Calculate the implied timescales for the given values.
    .. todo:: catch if for higher lagtimes the dimensionality changes

    Parameters
    ----------
    trajs : StateTraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    lagtimes : list or ndarray int
        Lagtimes for estimating the markov model given in [frames].
        This is not implemented yet!

    reversible : bool
        If reversibility should be enforced for the markov state model.

    Returns
    -------
    T : ndarray
        Transition rate matrix.

    """
    # format input
    trajs = StateTraj(trajs)
    lagtimes = np.atleast_1d(lagtimes)

    # check that lag times are array of integers
    if not np.issubdtype(lagtimes.dtype, np.integer):
        raise TypeError(
            'Lagtimes needs to be integers but are {0}'.format(lagtimes.dtype),
        )
    if not (lagtimes > 0).all():
        raise TypeError('Lagtimes needs to be positive integers')
    if reversible:
        raise TypeError('Reversible matrices are not anymore supported.')

    # initialize result
    impl_timescales = np.zeros((len(lagtimes), trajs.nstates - 1))

    for idx, lagtime in enumerate(lagtimes):
        transmat, _ = trajs.estimate_markov_model(lagtime)
        impl_timescales[idx] = _implied_timescales(transmat, lagtime)

    return impl_timescales


@decorit.alias('peq')
def equilibrium_population(tmat, allow_non_ergodic=True):
    """Calculate equilibirum population.

    If there are non ergodic states, their population is set to zero.

    Parameters
    ----------
    tmat : ndarray
        Quadratic transition matrix, needs to be ergodic.

    allow_non_ergodic : bool
        If True only the largest ergodic subset will be used. Otherwise it will
        throw an error if not ergodic.

    Returns
    -------
    peq : ndarray
        Equilibrium population of input matrix.

    """
    tmat = np.asarray(tmat)
    is_ergodic = tests.is_ergodic(tmat)
    if not allow_non_ergodic and not is_ergodic:
        raise ValueError('tmat needs to be ergodic transition matrix.')

    # calculate ev for ergodic subset
    if is_ergodic:
        _, eigenvectors = linalg.left_eigenvectors(tmat)
        eigenvectors = eigenvectors[0]
    else:
        mask = tests.ergodic_mask(tmat)
        _, evs_mask = linalg.left_eigenvectors(
            row_normalize_matrix(
                tmat[np.ix_(mask, mask)],
            ),
        )

        eigenvectors = np.zeros(len(tmat), dtype=tmat.dtype)
        eigenvectors[mask] = evs_mask[0]

    return eigenvectors / np.sum(eigenvectors)



# -*- coding: utf-8 -*-
"""Benchmarking tests.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest

import msmhelper as mh


@pytest.fixture
def state_traj():
    """Define state trajectory."""
    np.random.seed(137)
    return mh.StateTraj(np.random.randint(low=1, high=11, size=int(1e6)))


@pytest.fixture
def lagtime():
    """Define lag time."""
    return 1


def test_msm_msmhelper_statetraj(state_traj, lagtime, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(
        mh.msm.estimate_markov_model,
        state_traj,
        lagtime=lagtime,
    )


def test_msm_msmhelper_list(state_traj, lagtime, benchmark):
    """Benchmark msmhelper without StateTraj class."""
    state_traj = state_traj.state_trajs
    benchmark(
        mh.msm.estimate_markov_model,
        state_traj,
        lagtime=lagtime,
    )


def test_msm_pyemma(state_traj, lagtime, benchmark):
    """Benchmark pyemma without reversibility."""
    from pyemma import msm as emsm

    state_traj = state_traj.state_trajs
    benchmark(
        emsm.estimate_markov_model,
        state_traj,
        lagtime,
        reversible=False,
    )


def test_msm_pyemma_reversible(state_traj, lagtime, benchmark):
    """Benchmark pyemma with reversibility."""
    from pyemma import msm as emsm

    state_traj = state_traj.state_trajs
    benchmark(
        emsm.estimate_markov_model,
        state_traj,
        lagtime,
        reversible=True,
    )


def test_is_index_traj(state_traj, benchmark):
    """Test row normalization."""
    state_traj = state_traj.state_trajs
    benchmark(mh.tests.is_index_traj, state_traj)


def test_msm_wt(benchmark):
    np.random.seed(42)
    N = 50
    trajs = mh.StateTraj(np.random.randint(N, size=100))
    benchmark(
        mh.estimate_msm_waiting_times,
        trajs=trajs,
        lagtime=1,
        start=trajs.states.min(),
        final=trajs.states.max(),
        steps=int(1e8),
    )


def test_mcmc(benchmark):
    N = 50
    trajs = mh.StateTraj(np.random.randint(N, size=100))
    benchmark(
        mh.propagate_MCMC,
        trajs, 1, steps=int(1e7),
    )



# -*- coding: utf-8 -*-
"""Benchmarking Linalg.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest

import msmhelper as mh


@pytest.fixture
def tmat1():
    """Define transition matrix."""
    np.random.seed(137)
    tmat, _ = mh.estimate_markov_model(
        mh.StateTraj(np.random.randint(low=0, high=10, size=int(1e5))),
        lagtime=1,
    )
    return tmat


@pytest.fixture
def tmat2():
    """Define transition matrix."""
    np.random.seed(137)
    tmat, _ = mh.estimate_markov_model(
        mh.StateTraj(np.random.randint(low=0, high=100, size=int(1e5))),
        lagtime=1,
    )
    return tmat


@pytest.fixture
def tmat3():
    """Define transition matrix."""
    np.random.seed(137)
    tmat, _ = mh.estimate_markov_model(
        mh.StateTraj(np.random.randint(low=0, high=1000, size=int(1e5))),
        lagtime=1,
    )
    return tmat


# ~~~ TESTS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def test_linalg_eigvals_10states(tmat1, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(mh.linalg.eigvals, tmat1)


def test_linalg_eigvals_100states(tmat2, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(mh.linalg.eigvals, tmat2)


def test_linalg_eigvals_1000states(tmat3, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(mh.linalg.eigvals, tmat3)


def test_linalg_eig_10states(tmat1, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(mh.linalg.eig, tmat1)


def test_linalg_eig_100states(tmat2, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(mh.linalg.eig, tmat2)


def test_linalg_eig_1000states(tmat3, benchmark):
    """Benchmark msmhelper with StateTraj class."""
    benchmark(mh.linalg.eig, tmat3)



# -*- coding: utf-8 -*-
"""Benchmarking tests.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
# ~~~ IMPORT ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
import numpy as np
import pytest

import msmhelper as mh


@pytest.fixture
def traj1():
    """Define traj 1."""
    np.random.seed(137)
    return mh.StateTraj(np.random.randint(low=1, high=11, size=int(1e5)))


@pytest.fixture
def traj2():
    """Define traj 2."""
    np.random.seed(337)
    return mh.StateTraj(np.random.randint(low=1, high=11, size=int(1e5)))


# ~~~ TESTS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def test_compare_discretization(traj1, traj2, benchmark):
    """Benchmark msmhelper with pure numpy."""
    # execute for compiling numba
    mh.compare._compare_discretization(traj1, traj2, method='symmetric')
    benchmark(
        mh.compare._compare_discretization,
        traj1,
        traj2,
        method='symmetric',
    )



# -*- coding: utf-8 -*-
""".. include:: ../README.md"""

from .benchmark import (
    bh_test,
    buchete_hummer_test,
    chapman_kolmogorov_test,
    ck_test,
)
from .compare import compare_discretization
from .iotext import (
    opentxt,
    savetxt,
    opentxt_limits,
    openmicrostates,
    open_limits,
)
from .md import (
    estimate_msm_waiting_times,
    estimate_msm_wt,
    estimate_paths,
    estimate_waiting_times,
    estimate_wt,
    propagate_MCMC,
)
from .msm import (
    estimate_markov_model,
    equilibrium_population,
    implied_timescales,
    peq,
    row_normalize_matrix,
)
from .statetraj import LumpedStateTraj, StateTraj
from .tools import (
    rename_by_population,
    rename_by_index,
    runningmean,
    shift_data,
    unique,
)

__version__ = '0.6.1'



# -*- coding: utf-8 -*-
"""Set of helpful test functions.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import decorit
import numpy as np

from msmhelper import tools
from msmhelper.statetraj import StateTraj


def is_quadratic(matrix):
    """Check if matrix is quadratic.

    Parameters
    ----------
    matrix : ndarray, list of lists
        Matrix which is checked if is 2d array.

    Returns
    -------
    is_quadratic : bool

    """
    # cast to 2d for easier error checking

    matrix = np.atleast_2d(matrix)
    shape = np.shape(matrix)

    # Check whether matrix is quadratic.
    if shape[0] != shape[1]:
        return False
    # check if scalar or tensor higher than 2d
    if shape[0] == 1 or matrix.ndim > 2:
        return False

    return True


def is_state_traj(trajs):
    """Check if state trajectory is correct formatted.

    Parameters
    ----------
    trajs : list of ndarray
        State trajectory/trajectories need to be lists of ndarrays of integers.

    Returns
    -------
    is_state_traj : bool

    """
    try:
        tools._check_state_traj(trajs)  # noqa: WPS437
    except TypeError:
        return False
    else:
        return True


def is_index_traj(trajs):
    """Check if states can be used as indices.

    Parameters
    ----------
    trajs : list of ndarray
        State trajectory/trajectories need to be lists of ndarrays of integers.

    Returns
    -------
    is_index : bool

    """
    if isinstance(trajs, StateTraj):
        return True
    if is_state_traj(trajs):
        states = tools.unique(trajs)
        return np.array_equal(states, np.arange(len(states)))
    return False


@decorit.alias('is_tmat')
def is_transition_matrix(matrix, atol=1e-8):
    """Check if transition matrix.

    Rows and cols of zeros (non-visited states) are accepted.

    Parameters
    ----------
    matrix : ndarray
        Transition matrix.
    atol : float, optional
        Absolute tolerance.

    Returns
    -------
    is_tmat : bool

    """
    matrix = np.atleast_2d(matrix)
    visited = np.logical_or(
        matrix.sum(axis=-1),
        matrix.sum(axis=0),
    )
    return (
        is_quadratic(matrix) and
        np.logical_or(
            np.abs(matrix.sum(axis=-1) - 1) <= atol,
            ~visited,
        ).all()
    )


def is_ergodic(matrix, atol=1e-8):
    """Check if matrix is ergodic.

    Taken from:
    Wielandt, H. "Unzerlegbare, Nicht Negativen Matrizen."
    Mathematische Zeitschrift. Vol. 52, 1950, pp. 642–648.

    Parameters
    ----------
    matrix : ndarray
        Transition matrix.
    atol : float, optional
        Absolute tolerance.

    Returns
    -------
    is_ergodic : bool

    """
    if not is_transition_matrix(matrix):
        return False

    matrix = np.atleast_2d(matrix)

    nstates = len(matrix)
    exponent = (nstates - 1)**2 + 1

    matrix = tools.matrix_power(matrix, exponent)
    return (matrix > atol).all()


def is_fuzzy_ergodic(matrix, atol=1e-8):
    """Check if matrix is ergodic, up to missing states or trap states.

    If there are two or more disjoint

    Parameters
    ----------
    matrix : ndarray
        Transition matrix.
    atol : float, optional
        Absolute tolerance.

    Returns
    -------
    is_fuzzy_ergodic : bool

    """
    if not is_transition_matrix(matrix):
        return False

    matrix = np.atleast_2d(matrix)
    row_col_sum = matrix.sum(axis=-1) + matrix.sum(axis=0)

    is_trap_state = np.logical_or(
        np.abs(row_col_sum - 2) <= atol,
        np.abs(row_col_sum) <= atol,
    )
    is_trap_state = np.logical_or(
        is_trap_state[:, np.newaxis],
        is_trap_state[np.newaxis, :],
    )

    nstates = len(matrix)
    exponent = (nstates - 1)**2 + 1
    matrix = tools.matrix_power(matrix, exponent)

    return np.logical_or(matrix > 0, is_trap_state).all()


def ergodic_mask(matrix, atol=1e-8):
    """Create mask for filtering ergodic submatrix.

    Parameters
    ----------
    matrix : ndarray
        Transition matrix.
    atol : float, optional
        Absolute tolerance.

    Returns
    -------
    mask : bool ndarray

    """
    if not is_transition_matrix(matrix):
        raise ValueError("Input matrix needs to be of kind transition matrix.")

    matrix = np.atleast_2d(matrix)
    nstates = len(matrix)
    exponent = (nstates - 1)**2 + 1

    matrix = tools.matrix_power(matrix, exponent) > atol
    matrix = np.logical_and(matrix, matrix.T)

    # find minimum row counts to identify largest connected block
    maxstates = np.max(matrix.sum(axis=-1))
    return matrix.sum(axis=-1) == maxstates



# -*- coding: utf-8 -*-
"""Set of helpful functions.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

TODO:
    - Correct border effects of running mean

"""
import datetime
import getpass  # noqa: SC100 # get user name with getpass.getuser()
import os
import platform  # get pc name with platform.node()
import sys

import numba
import numpy as np

import __main__ as main
from msmhelper.statetraj import StateTraj


def shift_data(array, val_old, val_new, dtype=np.int64):
    """Shift integer array (data) from old to new values.

    > **CAUTION:**
    > The values of `val_old`, `val_new` and `data` needs to be integers.

    The basic function is based on Ashwini_Chaudhary solution:
    https://stackoverflow.com/a/29408060

    Parameters
    ----------
    array : StateTraj or ndarray or list or list of ndarrays
        1D data or a list of data.

    val_old : ndarray or list
        Values in data which should be replaced. All values needs to be within
        the range of `[data.min(), data.max()]`

    val_new : ndarray or list
        Values which will be used instead of old ones.

    dtype : data-type, optional
        The desired data-type. Needs to be of type unsigned integer.

    Returns
    -------
    array : ndarray
        Shifted data in same shape as input.

    """
    # check data-type
    if not np.issubdtype(dtype, np.integer):
        raise TypeError('An unsigned integer type is needed.')

    # flatten data
    array, shape_kwargs = _flatten_data(array)

    # offset data and val_old to allow negative values
    offset = np.min([np.min(array), np.min(val_new)])

    # convert to np.array
    val_old = (np.asarray(val_old) - offset).astype(dtype)
    val_new = (np.asarray(val_new) - offset).astype(dtype)

    # convert data and shift
    array = (array - offset).astype(dtype)

    # shift data
    conv = np.arange(array.max() + 1, dtype=dtype)
    conv[val_old] = val_new
    array = conv[array]

    # shift data back
    array = array.astype(np.int32) + offset

    # reshape and return
    return _unflatten_data(array, shape_kwargs)


def rename_by_population(trajs, return_permutation=False):
    r"""Rename states sorted by their population starting from 1.

    Parameters
    ----------
    trajs : list or ndarray or list of ndarrays
        State trajectory or list of state trajectories.

    return_permutation : bool
        Return additionaly the permutation to achieve performed renaming.
        Default is False.

    Returns
    -------
    trajs : ndarray
        Renamed data.

    permutation : ndarray
        Permutation going from old to new state nameing. So the `i`th state
        of the new naming corresponds to the old state `permutation[i-1]`.

    """
    # get unique states with population
    states, pop = unique(trajs, return_counts=True)

    # get decreasing order
    idx_sort = np.argsort(pop)[::-1]
    states = states[idx_sort]

    # rename states
    trajs_renamed = shift_data(
        trajs,
        val_old=states,
        val_new=np.arange(len(states)) + 1,
    )
    if return_permutation:
        return trajs_renamed, states
    return trajs_renamed


def rename_by_index(trajs, return_permutation=False):
    r"""Rename states sorted by their numerical values starting from 0.

    Parameters
    ----------
    trajs : list or ndarray or list of ndarrays
        State trajectory or list of state trajectories.

    return_permutation : bool
        Return additionaly the permutation to achieve performed renaming.
        Default is False.

    Returns
    -------
    trajs : ndarray
        Renamed data.

    permutation : ndarray
        Permutation going from old to new state nameing. So the `i`th state
        of the new naming corresponds to the old state `permutation[i-1]`.

    """
    # get unique states
    states = unique(trajs)

    # rename states
    trajs_renamed = shift_data(
        trajs,
        val_old=states,
        val_new=np.arange(len(states)),
    )
    if return_permutation:
        return trajs_renamed, states
    return trajs_renamed


def unique(trajs, **kwargs):
    r"""Apply numpy.unique to traj.

    Parameters
    ----------
    trajs : list or ndarray or list of ndarrays
        State trajectory or list of state trajectories.

    kwargs
        Arguments of [numpy.unique()](NP_DOC.numpy.unique.html)

    Returns
    -------
    unique
        Array containing all states, see numpy for more details.

    """
    # flatten data
    trajs, _ = _flatten_data(trajs)

    # get unique states with population
    return np.unique(trajs, **kwargs)


def runningmean(array, window):
    r"""Compute centered running average with given window size.

    This function returns the centered based running average of the given
    data. The output of this function is of the same length as the input,
    by assuming that the given data is zero before and after the given
    series. Hence, there are border affects which are not corrected.

    > **CAUTION:**
    > If the given window is even (not symmetric) it will be shifted towards
    > the beginning of the current value. So for `window=4`, it will consider
    > the current position \(i\), the two to the left \(i-2\) and \(i-1\) and
    > one to the right \(i+1\).

    Function is taken from lapis:
    https://stackoverflow.com/questions/13728392/moving-average-or-running-mean

    Parameters
    ----------
    array : ndarray
        One dimensional numpy array.

    window : int
        Integer which specifies window-width.

    Returns
    -------
    array_rmean : ndarray
        Data which is time-averaged over the specified window.

    """
    # Calculate running mean
    return np.convolve(
        array,
        np.ones(window) / window,
        mode='same',
    )


def swapcols(array, indicesold, indicesnew):
    r"""Interchange cols of an ndarray.

    This method swaps the specified columns.
    .. todo:: Optimize memory usage

    Parameters
    ----------
    array : ndarray
        2D numpy array.
    indicesold : integer or ndarray
        1D array of indices.
    indicesnew : integer or ndarray
        1D array of new indices

    Returns
    -------
    array_swapped : ndarray
        2D numpy array with swappend columns.

    """
    # cast to 1d arrays
    indicesnew = _asindex(indicesnew)
    indicesold = _asindex(indicesold)

    if len(indicesnew) != len(indicesold):
        raise ValueError('Indices needs to be of same shape.')

    # cast data
    array = np.asarray(array)

    if np.all(indicesnew == indicesold):
        return array

    # fails for large data sets
    # noqa: E800 # array.T[indicesold] = array.T[indicesnew]
    array_swapped = np.copy(array)
    array_swapped.T[indicesold] = array.T[indicesnew]

    return array_swapped


def get_runtime_user_information():
    r"""Get user runtime information.

    > **CAUTION:**
    > For python 3.5 or lower the date is not formatted and contains
    > microscends.

    Returns
    -------
    RUI : dict
        Holding username in 'user', pc name in 'pc', date of execution 'date',
        path of execution 'script_dir' and name of execution main file
        'script_name'. In case of interactive usage, script_name is 'console'.

    """
    try:
        script_dir, script_name = os.path.split(
            os.path.abspath(main.__file__),  # noqa: WPS609
        )
    except AttributeError:  # pragma: no cover
        script_dir, script_name = '', 'console'

    # get time without microseconds
    date = datetime.datetime.now()
    if sys.version_info >= (3, 6):
        date = date.isoformat(sep=' ', timespec='seconds')

    return {
        'user': getpass.getuser(),
        'pc': platform.node(),
        'date': date,
        'script_dir': script_dir,
        'script_name': script_name,
    }


def _asindex(idx):
    """Cast to 1d integer ndarray."""
    idx = np.atleast_1d(idx).astype(np.int64)
    if len(idx.shape) > 1:
        raise ValueError('Wrong dimensionality of indices.')
    return idx


def format_state_traj(trajs):
    """Convert state trajectory to list of ndarrays.

    Parameters
    ----------
    trajs : list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    Returns
    -------
    trajs : list of ndarray
        Return list of ndarrays of integers.

    """
    # list or tuple
    if isinstance(trajs, (tuple, list)):
        # list of integers
        if all((np.issubdtype(type(state), np.integer) for state in trajs)):
            trajs = [np.array(trajs)]
        # list of lists
        elif all((isinstance(traj, list) for traj in trajs)):
            trajs = [np.array(traj) for traj in trajs]
    # ndarray
    if isinstance(trajs, np.ndarray):
        if len(trajs.shape) == 1:
            trajs = [trajs]
        elif len(trajs.shape) == 2:
            trajs = list(trajs)

    # check for integers
    _check_state_traj(trajs)

    return trajs


def _check_state_traj(trajs):
    """Check if state trajectory is correct formatted."""
    # check for integers
    if isinstance(trajs, StateTraj):
        return True

    for traj in trajs:
        if not isinstance(traj, np.ndarray):
            raise TypeError(
                'Trajs need to be np.ndarray but are {0}'.format(type(traj)),
            )
        if not np.issubdtype(traj.dtype, np.integer):
            raise TypeError(
                'States needs to be integers but are {0}'.format(traj.dtype),
            )
    return True


def _flatten_data(array):
    """Flatten data to 1D ndarray.

    This method flattens ndarrays, list of ndarrays to a 1D ndarray. This can
    be undone with _unflatten_data().

    Parameters
    ----------
    array : list or ndarray or list of ndarrays
        1D data or a list of data.

    Returns
    -------
    data : ndarray
        Flattened data.

    kwargs : dict
        Dictionary with information to restore shape.

    """
    kwargs = {}

    # flatten data
    if isinstance(array, (tuple, list)):
        # list of ndarrays, lists or tuples
        if all((isinstance(row, (np.ndarray, tuple, list)) for row in array)):
            # get shape and flatten
            kwargs['limits'] = np.cumsum([len(row) for row in array])
            array = np.concatenate(array)
        # list of numbers
        else:
            array = np.asarray(array)
    elif isinstance(array, np.ndarray):
        # get shape and flatten
        kwargs['data_shape'] = array.shape
        array = array.flatten()

    return array, kwargs


def _unflatten_data(array, kwargs):
    """Unflatten data to original structure.

    This method undoes _flatten_data().

    Parameters
    ----------
    array : ndarray
        Flattened data.

    kwargs : dict
        Dictionary with information to restore shape. Provided by
        _flatten_data().

    Returns
    -------
    array : ndarray, list, list of ndarrays
        Data with restored shape.

    """
    # parse kwargs
    data_shape = kwargs.get('data_shape')
    limits = kwargs.get('limits')

    # reshape
    if data_shape is not None:
        array = array.reshape(data_shape)
    elif limits is not None:
        array = np.split(array, limits)[:-1]

    return array


@numba.njit
def matrix_power(matrix, power):
    """Calculate matrix power with np.linalg.matrix_power.

    Same as numpy function, except only for float matrices. See
    [np.linalg.matrix_power](NP_DOC/numpy.linalg.matrix_power.html).

    Parameters
    ----------
    matrix : ndarray
        2d matrix of type float.

    power : int, float
        Power of matrix.

    Returns
    -------
    matpow : ndarray
        Matrix power.

    """
    return np.linalg.matrix_power(matrix, power)


@numba.njit
def find_first(search_val, array):
    """Return first occurance of item in array."""
    for idx, idx_val in enumerate(array):
        if search_val == idx_val:
            return idx
    return -1



# -*- coding: utf-8 -*-
"""Basic linear algebra method.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import decorit
import numpy as np

from msmhelper import tests


@decorit.alias('eigl')
def left_eigenvectors(matrix):
    """Estimate left eigenvectors.

    Estimates the left eigenvectors and corresponding eigenvalues of a
    quadratic matrix.

    Parameters
    ----------
    matrix : ndarray
        Quadratic 2d matrix eigenvectors and eigenvalues or determined of.

    Returns
    -------
    eigenvalues : ndarray
        N eigenvalues sorted by their value (descending).

    eigenvectors : ndarray
        N eigenvectors sorted by descending eigenvalues.

    """
    matrix = np.asarray(matrix)
    if not tests.is_quadratic(matrix):
        raise TypeError('Matrix needs to be quadratic {0}'.format(matrix))

    eigenvalues, eigenvectors = _eigenvectors(matrix.transpose())
    return np.real_if_close(eigenvalues), np.real_if_close(eigenvectors)


@decorit.alias('eig')
def right_eigenvectors(matrix):
    """Estimate right eigenvectors.

    Estimates the right eigenvectors and corresponding eigenvalues of a
    quadratic matrix.

    Parameters
    ----------
    matrix : ndarray
        Quadratic 2d matrix eigenvectors and eigenvalues or determined of.

    Returns
    -------
    eigenvalues : ndarray
        N eigenvalues sorted by their value (descending).

    eigenvectors : ndarray
        N eigenvectors sorted by descending eigenvalues.

    """
    matrix = np.asarray(matrix)
    if not tests.is_quadratic(matrix):
        raise TypeError('Matrix needs to be quadratic {0}'.format(matrix))

    eigenvalues, eigenvectors = _eigenvectors(matrix)
    return np.real_if_close(eigenvalues), np.real_if_close(eigenvectors)


def _eigenvectors(matrix):
    """Estimate eigenvectors."""
    eigenvalues, eigenvectors = np.linalg.eig(matrix)

    # Transpose eigenvectors, since v[:,i] is eigenvector
    eigenvectors = eigenvectors.T

    # Sort them by descending eigenvalues
    idx_eigenvalues = eigenvalues.argsort()[::-1]

    return eigenvalues[idx_eigenvalues], eigenvectors[idx_eigenvalues]


@decorit.alias('eiglvals')
def left_eigenvalues(matrix):
    """Estimate left eigenvalues.

    Estimates the left eigenvalues of a quadratic matrix.

    Parameters
    ----------
    matrix : ndarray
        Quadratic 2d matrix eigenvalues or determined of.

    Returns
    -------
    eigenvalues : ndarray
        N eigenvalues sorted by their value (descending).

    """
    matrix = np.asarray(matrix)
    if not tests.is_quadratic(matrix):
        raise TypeError('Matrix needs to be quadratic {0}'.format(matrix))

    return np.real_if_close(_eigenvalues(np.transpose(matrix)))


@decorit.alias('eigvals')
def right_eigenvalues(matrix):
    """Estimate right eigenvalues.

    Estimates the right eigenvalues of a quadratic matrix.

    Parameters
    ----------
    matrix : ndarray
        Quadratic 2d matrix eigenvalues or determined of.

    Returns
    -------
    eigenvalues : ndarray
        N eigenvalues sorted by their value (descending).

    """
    matrix = np.asarray(matrix)
    if not tests.is_quadratic(matrix):
        raise TypeError('Matrix needs to be quadratic {0}'.format(matrix))

    return np.real_if_close(_eigenvalues(matrix))


def _eigenvalues(matrix):
    """Estimate eigenvalues."""
    eigenvalues = np.linalg.eigvals(matrix)

    # Sort them by descending eigenvalues
    idx_eigenvalues = eigenvalues.argsort()[::-1]

    return eigenvalues[idx_eigenvalues]



# -*- coding: utf-8 -*-
"""Set of helpful functions for comparing markov state models.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numba
import numpy as np

from msmhelper.statetraj import StateTraj


def compare_discretization(traj1, traj2, method='symmetric'):
    """Compare similarity of two state discretizations.

    This method compares the similarity of two state discretizations of the
    same dataset. There are two different methods, 'directed' gives a measure
    on how high is the probable to assign a frame correclty knowing the
    `traj1`. Hence splitting a state into many is not penalized, while merging
    multiple into a single state is. Selecting 'symmetric' it is check in both
    directions, so it checks for each state if it is possible to assigned it
    forward or backward. Hence, splitting and merging states is not penalized.

    Parameters
    ----------
    traj1 : StateTraj like
        First state discretization.
    traj2 : StateTraj like
        Second state discretization.
    method : ['symmetric', 'directed']
        Selecting similarity norm. 'symmetric' compares if each frame is
        forward or backward assignable, while 'directed' checks only if it is
        forard assignable.

    Returns
    -------
    similarity : float
        Similarity going from [0, 1], where 1 means identical and 0 no
        similarity at all.

    """
    # format input
    traj1, traj2 = StateTraj(traj1), StateTraj(traj2)
    if method not in {'symmetric', 'directed'}:
        raise ValueError(
            'Only methods "symmetric" and "directed" are supported',
        )

    # check if same length
    if traj1.nframes != traj2.nframes:
        raise ValueError(
            'Trajectories are of different length: ' +
            '{0} vs {1}'.format(traj1.nframes, traj2.nframes),
        )

    # check if only single state
    if traj1.nstates == 1 or traj2.nstates == 1:
        raise ValueError(
            'Trajectories needs to have at least two states: ' +
            '{0} and {1}'.format(traj1.nstates, traj2.nstates),
        )
    return _compare_discretization(traj1, traj2, method)


def _compare_discretization(traj1, traj2, method):
    """Compare similarity of two state discretizations."""
    traj1_flat, traj2_flat = traj1.trajs_flatten, traj2.trajs_flatten
    idx1 = [
        np.where(traj1_flat == state)[0] for state in range(traj1.nstates)
    ]
    idx2 = [
        np.where(traj2_flat == state)[0] for state in range(traj2.nstates)
    ]
    if not numba.config.DISABLE_JIT:  # pragma: no cover
        idx1 = numba.typed.List(idx1)
        idx2 = numba.typed.List(idx2)

    # intersect arrays and normalize to length of lists
    intersect = _intersect_array(idx1, idx2)
    intersect12 = intersect / np.array(
        [len(idx) for idx in idx1],
    )[:, np.newaxis]
    intersect21 = intersect.T / np.array(
        [len(idx) for idx in idx2],
    )[:, np.newaxis]

    if method == 'symmetric':
        return _compare_trajs_symmetric(
            traj1_flat, traj2_flat, intersect12, intersect21,
        )
    elif method == 'directed':
        return _compare_trajs_directed(
            traj1_flat, traj2_flat, intersect12, intersect21,
        )
    raise ValueError('This should never be reached')


@numba.njit(parallel=True)
def _compare_trajs_symmetric(traj1, traj2, intersect12, intersect21):
    nframes = len(traj1)
    similarity = 0
    for idx in numba.prange(nframes):
        state1, state2 = traj1[idx], traj2[idx]
        similarity += max([
            intersect12[state1, state2], intersect21[state2, state1],
        ])
    return similarity / nframes


@numba.njit(parallel=True)
def _compare_trajs_directed(traj1, traj2, intersect12, intersect21):
    nframes = len(traj1)
    similarity = 0
    for idx in numba.prange(nframes):
        state1, state2 = traj1[idx], traj2[idx]
        similarity += intersect21[state2, state1]
    return similarity / nframes


@numba.njit(parallel=True)
def _intersect_array(idx1, idx2):
    """Intersect two list of arrays."""
    len1, len2 = len(idx1), len(idx2)
    intersect = np.empty((len1, len2), dtype=np.float64)

    for i in range(len1):
        for j in numba.prange(len2):
            intersect[i, j] = _intersect(idx1[i], idx2[j])

    return intersect


@numba.njit
def _intersect(ar1, ar2):
    """Intersect unique sorted list ar1 with ar2."""
    idx1, idx2 = 0, 0
    len1, len2 = len(ar1), len(ar2)

    count = 0
    while idx1 < len1 and idx2 < len2:
        if ar1[idx1] == ar2[idx2]:
            count += 1
            idx1 += 1
            idx2 += 1
        elif ar1[idx1] > ar2[idx2]:
            idx2 += 1
        else:
            idx1 += 1

    return count



# -*- coding: utf-8 -*-
"""Tests for the md module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest

import msmhelper


@pytest.mark.parametrize('traj, start, final, reftimes', [
    ([1, 2, 1, 2, 1, 1, 3, 4, 3, 2, 1, 2, 1, 4, 3, 4], 1, 3, [6, 4]),
    ([1, 2, 1, 2, 1, 1, 3, 4, 3, 2, 1, 2, 1, 4, 3, 4], [1, 2], [3, 4], [6, 4]),
])
def test_estimate_waiting_times(traj, start, final, reftimes):
    """Test estimate waiting time."""
    times = msmhelper.estimate_waiting_times(traj, start, final)
    np.testing.assert_array_equal(times, reftimes)

    # check overlapping start and final states
    with pytest.raises(ValueError):
        msmhelper.estimate_waiting_times(traj, start, start)

    # check state not contained in traj
    with pytest.raises(ValueError):
        msmhelper.estimate_waiting_times(traj, start, np.max(traj) + 1)


@pytest.mark.parametrize('traj, start, final, refpaths', [
    (
        [1, 2, 1, 2, 1, 1, 3, 4, 3, 2, 1, 2, 1, 4, 3, 4],
        1,
        3,
        {(1, 3): [6], (1, 4, 3): [4]},
    ),
    (
        [1, 2, 1, 2, 1, 1, 3, 4, 3, 2, 1, 2, 1, 4, 3, 4],
        [1, 2],
        [3, 4],
        {(1, 3): [6], (1, 4): [4]},
    ),
    (
        [1, 2, 3, 4, 5, 2, 6],
        1,
        6,
        {(1, 2, 6): [6]},
    ),
])
def test_estimate_paths(traj, start, final, refpaths):
    """Test estimate waiting time."""
    paths = msmhelper.estimate_paths(traj, start, final)
    assert paths == refpaths

    # check overlapping start and final states
    with pytest.raises(ValueError):
        msmhelper.estimate_paths(traj, start, start)

    # check state not contained in traj
    with pytest.raises(ValueError):
        msmhelper.estimate_paths(traj, start, np.max(traj) + 1)



# -*- coding: utf-8 -*-
"""Class for handling discrete state trajectories.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np

import msmhelper as mh


class StateTraj:  # noqa: WPS214
    """Class for handling discrete state trajectories."""

    # noqa: E800 # add slots magic,  self.__slots__ = ('_states', '_trajs').
    def __new__(cls, trajs):
        """Initialize new instance.

        If called with instance, return instance instead.

        """
        if isinstance(trajs, StateTraj):
            return trajs
        return super().__new__(cls)

    def __init__(self, trajs):
        """Initialize StateTraj and convert to index trajectories.

        If called with StateTraj instance, it will be retuned instead.

        Parameters
        ----------
        trajs : list or ndarray or list of ndarray
            State trajectory/trajectories. The states need to be integers.

        """
        if isinstance(trajs, StateTraj):
            return

        self.state_trajs = trajs

    @property
    def states(self):
        """Return active set of states.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return self._states

    @property
    def nstates(self):
        """Return number of states.

        Returns
        -------
        nstates : int
            Number of states.

        """
        return len(self.states)

    @property
    def ntrajs(self):
        """Return number of trajectories.

        Returns
        -------
        ntrajs : int
            Number of trajectories.

        """
        return len(self)

    @property
    def nframes(self):
        """Return cummulated length of all trajectories.

        Returns
        -------
        nframes : int
            Number of frames of all trajectories.

        """
        return self._nframes

    @property
    def state_trajs(self):
        """Return state trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        if np.array_equal(self.states, np.arange(self.nstates)):
            return self.trajs
        return mh.shift_data(
            self.trajs,
            np.arange(self.nstates),
            self.states,
        )

    @state_trajs.setter
    def state_trajs(self, trajs):
        """Set the state trajectory.

        Parameters
        ----------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        self._trajs = mh.tools.format_state_traj(trajs)

        # get number of states
        self._states = mh.tools.unique(self._trajs)

        # shift to indices
        if not np.array_equal(self._states, np.arange(self.nstates)):
            self._trajs, self._states = mh.tools.rename_by_index(
                self._trajs,
                return_permutation=True,
            )

        # set number of frames
        self._nframes = np.sum([len(traj) for traj in self.trajs])

    @property
    def state_trajs_flatten(self):
        """Return flattened state trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarrays representation of state trajectories.

        """
        return np.concatenate(self.state_trajs)

    @property
    def index_trajs(self):
        """Return index trajectory.

        Same as `self.trajs`

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        return self.trajs

    @property
    def trajs(self):
        """Return index trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        return self._trajs

    @property
    def trajs_flatten(self):
        """Return flattened index trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarrays representation of index trajectories.

        """
        return np.concatenate(self._trajs)

    def __repr__(self):
        """Return representation of class."""
        kw = {
            'clname': self.__class__.__name__,
            'trajs': self.state_trajs,
        }
        return ('{clname}({trajs})'.format(**kw))

    def __str__(self):
        """Return string representation of class."""
        return ('{trajs!s}'.format(trajs=self.state_trajs))

    def __iter__(self):
        """Iterate over trajectories."""
        return iter(self.state_trajs)

    def __len__(self):
        """Return length of list of trajectories."""
        return len(self.trajs)

    def __getitem__(self, key):
        """Get key value."""
        return self.state_trajs.__getitem__(key)  # noqa: WPS609

    def __eq__(self, other):
        """Compare two objects."""
        if not isinstance(other, StateTraj):
            return NotImplemented
        return (
            self.ntrajs == other.ntrajs and
            all(
                np.array_equal(self[idx], other[idx])
                for idx in range(self.ntrajs)
            )
        )

    def estimate_markov_model(self, lagtime):
        """Estimates Markov State Model.

        This method estimates the MSM based on the transition count matrix.

        Parameters
        ----------
        lagtime : int
            Lag time for estimating the markov model given in [frames].

        Returns
        -------
        T : ndarray
            Transition rate matrix.

        permutation : ndarray
            Array with corresponding states.

        """
        return mh.estimate_markov_model(self, lagtime)

    def state_to_idx(self, state):
        """Get idx corresponding to state.

        Parameters
        ----------
        state : int
            State to get idx of.

        Returns
        -------
        idx : int
            Idx corresponding to state.

        """
        idx = np.where(self.states == state)[0]
        if not idx.size:
            raise ValueError(
                'State "{state}" does not exists in trajectory.'.format(
                    state=state,
                ),
            )
        return idx[0]


class LumpedStateTraj(StateTraj):
    """Class for handling lumped discrete state trajectories."""

    def __new__(cls, macrotrajs, microtrajs=None):
        """Initialize new instance."""
        if isinstance(macrotrajs, LumpedStateTraj):
            return macrotrajs
        return super().__new__(cls, None)

    def __init__(self, macrotrajs, microtrajs=None):
        """Initialize LumpedStateTraj and convert to index trajectories.

        If called with LumpedStateTraj instance, it will be retuned instead.

        Parameters
        ----------
        macrotrajs : list or ndarray or list of ndarray
            Lumped state trajectory/trajectories. The states need to be
            integers and all states needs to correspond to union of
            microstates.

        microtrajs : list or ndarray or list of ndarray
            State trajectory/trajectories. EaThe states should start from zero
            and need to be integers.

        """
        if isinstance(macrotrajs, LumpedStateTraj):
            return

        if microtrajs is None:
            raise TypeError(
                'microtrajs may only be None when macrotrajs is of type ' +
                'LumpedStateTraj.',
            )
        # initialize base class
        self._parse_microtrajs(microtrajs)

        # parse the microstate to macrostate lumping
        self._parse_macrotrajs(macrotrajs)

    @property
    def states(self):
        """Return active set of macrostates.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return self._macrostates

    @property
    def nstates(self):
        """Return number of macrostates.

        Returns
        -------
        nstates : int
            Number of states.

        """
        return len(self.states)

    @property
    def microstate_trajs(self):
        """Return microstate trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        if np.array_equal(self.microstates, np.arange(self.nmicrostates)):
            return self.trajs
        return mh.shift_data(
            self.trajs,
            np.arange(self.nmicrostates),
            self.microstates,
        )

    @property
    def microstate_trajs_flatten(self):
        """Return flattened state trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarrays representation of state trajectories.

        """
        return np.concatenate(self.microstate_trajs)

    @property
    def state_trajs(self):
        """Return macrostate trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input macrostate data.

        """
        return mh.shift_data(
            self.trajs,
            np.arange(self.nmicrostates),
            self.state_assignment,
        )

    @property
    def state_trajs_flatten(self):
        """Return flattened macrostate trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarrays representation of macrostate trajectories.

        """
        return np.concatenate(self.state_trajs)

    @property
    def microstates(self):
        """Return active set of states.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return self._states

    @property
    def nmicrostates(self):
        """Return number of states.

        Returns
        -------
        nstates : int
            Number of states.

        """
        return len(self.microstates)

    def __repr__(self):
        """Return representation of class."""
        kw = {
            'clname': self.__class__.__name__,
            'trajs': self.state_trajs,
            'microtrajs': self.microstate_trajs,
        }
        return ('{clname}({trajs}, {microtrajs})'.format(**kw))

    def __str__(self):
        """Return string representation of class."""
        return ('{trajs!s}'.format(trajs=self.state_trajs))

    def __iter__(self):
        """Iterate over trajectories."""
        return iter(self.state_trajs)

    def __len__(self):
        """Return length of list of trajectories."""
        return len(self.trajs)

    def __getitem__(self, key):
        """Get key value."""
        return self.state_trajs.__getitem__(key)  # noqa: WPS609

    def __eq__(self, other):
        """Compare two objects."""
        if not isinstance(other, LumpedStateTraj):
            return NotImplemented
        return (
            self.ntrajs == other.ntrajs and
            all(
                np.array_equal(self.trajs[idx], other.trajs[idx])
                for idx in range(self.ntrajs)
            ) and
            np.array_equal(self.state_assignment, other.state_assignment)
        )

    def estimate_markov_model(self, lagtime):
        """Estimates Markov State Model.

        This method estimates the microstate MSM based on the transition count
        matrix, followed by Szabo-Hummer projection formalism to macrostates.

        Parameters
        ----------
        lagtime : int
            Lag time for estimating the markov model given in [frames].

        Returns
        -------
        T : ndarray
            Transition rate matrix.

        permutation : ndarray
            Array with corresponding states.

        """
        # in the following corresponds 'i' to micro and 'a' to macro
        msm_i, _ = mh.estimate_markov_model(self, lagtime)
        if not mh.tests.is_ergodic(msm_i):
            raise TypeError('tmat needs to be ergodic transition matrix.')
        return (self._estimate_markov_model(msm_i), self.states)

    def _estimate_markov_model(self, msm_i):
        """Estimates Markov State Model."""
        ones_i = np.ones_like(self.microstates)
        ones_a = np.ones_like(self.states)
        id_i = np.diag(ones_i)
        id_a = np.diag(ones_a)

        peq_i = mh.peq(msm_i)
        peq_a = np.array([
            np.sum(peq_i[self.state_assignment == state])
            for state in self.states
        ])
        d_i = np.diag(peq_i)
        d_a = np.diag(peq_a)
        aggret = np.zeros((self.nmicrostates, self.nstates))
        aggret[(np.arange(self.nmicrostates), self.state_assignment_idx)] = 1

        m_prime = np.linalg.inv(
            id_i +
            ones_i[:, np.newaxis] * peq_i[np.newaxis:, ] -
            msm_i,
        )
        m_twoprime = np.linalg.inv(
            np.linalg.multi_dot((aggret.T, d_i, m_prime, aggret)),
        )

        msm_a = (
            id_a +
            ones_a[:, np.newaxis] * peq_a[np.newaxis:, ] -
            m_twoprime @ d_a
        )
        return mh.row_normalize_matrix(msm_a)

    def _parse_macrotrajs(self, macrotrajs):
        """Parse the macrotrajs."""
        # TODO: improve performance by not using StateTraj
        macrotrajs = StateTraj(macrotrajs)
        self._macrostates = macrotrajs.states.copy()

        # cache flattened trajectories to speed up code for many states
        macrotrajs_flatten = macrotrajs.state_trajs_flatten
        microtrajs_flatten = self.microstate_trajs_flatten

        self.state_assignment = np.zeros(self.nmicrostates, dtype=np.int64)
        for idx, microstate in enumerate(self.microstates):
            idx_first = mh.tools.find_first(microstate, microtrajs_flatten)
            self.state_assignment[idx] = macrotrajs_flatten[idx_first]

        self.state_assignment_idx = mh.shift_data(
            self.state_assignment,
            self.states,
            np.arange(self.nstates),
        )

    def _parse_microtrajs(self, trajs):
        """Parse the microtrajs."""
        self._trajs = mh.tools.format_state_traj(trajs)

        # get number of states
        self._states = mh.tools.unique(self._trajs)

        # shift to indices
        if not np.array_equal(self._states, np.arange(self.nmicrostates)):
            self._trajs, self._states = mh.tools.rename_by_index(
                self._trajs,
                return_permutation=True,
            )

        # set number of frames
        self._nframes = np.sum([len(traj) for traj in self.trajs])



# -*- coding: utf-8 -*-
"""Benchmark Markov State Model.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import decorit
import numba
import numpy as np

from msmhelper import linalg, msm, tests, tools
from msmhelper.statetraj import LumpedStateTraj, StateTraj


@decorit.alias('ck_test')
def chapman_kolmogorov_test(trajs, lagtimes, tmax):
    r"""Calculate the Chapman Kolmogorov equation.

    This method estimates the Chapman Kolmogorov equation
    $$T(\tau n) = T^n(\tau)\;.$$
    Projected onto the diagonal this is known as the Chapman Kolmogorov test.
    For more details see, e.g., the review Prinz et al.[^1].

    [^1]: Prinz et al.
        **Markov models of molecular kinetics: Generation and validation**,
        *J. Chem. Phys.*, 134, 174105 (2011),
        doi:[10.1063/1.3565032](https://doi.org/10.1063/1.3565032)

    Parameters
    ----------
    trajs : StateTraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    lagtimes : list or ndarray int
        Lagtimes for estimating the markov model given in [frames].

    tmax : int
        Longest time to evaluate the CK equation given in [frames].

    Returns
    -------
    cktest : dict
        Dictionary holding for each lagtime the ckequation and with 'md' the
        reference.

    """
    # format input
    trajs = StateTraj(trajs)
    lagtimes = np.atleast_1d(lagtimes)
    lagtimes = np.sort(lagtimes)

    # check that lag times are array of integers
    if not np.issubdtype(lagtimes.dtype, np.integer):
        raise TypeError(
            'Lagtimes needs to be integers but are {0}'.format(lagtimes.dtype),
        )
    if not (lagtimes > 0).all():
        raise TypeError('Lagtimes needs to be positive integers')

    if lagtimes.ndim != 1:
        raise TypeError(
            'Lagtimes needs to be maximal 1d, but {0}'.format(lagtimes),
        )

    if not isinstance(tmax, int) or tmax < 0:
        raise TypeError('tmax needs to be a positive integer')

    ckeqs = {}
    for lagtime in lagtimes:
        ckeqs[lagtime] = _chapman_kolmogorov_test(trajs, lagtime, tmax)
    ckeqs['md'] = _chapman_kolmogorov_test_md(
        trajs, tmin=lagtimes[0], tmax=tmax,
    )

    return ckeqs


def _chapman_kolmogorov_test(trajs, lagtime, tmax):
    r"""Calculate the Chapman Kolmogorov equation $T^n(\tau)$."""
    times = _calc_times(lagtime=lagtime, tmax=tmax)
    ntimes = len(times)
    ckeq = np.empty((trajs.nstates, ntimes))

    # estimate Markov model
    tmat, _ = trajs.estimate_markov_model(lagtime=lagtime)

    is_ergodic = tests.is_ergodic(tmat)
    is_fuzzy_ergodic = tests.is_fuzzy_ergodic(tmat)
    for idx in range(ntimes):
        tmatpow = tools.matrix_power(tmat, idx + 1)
        ckeq[:, idx] = np.diagonal(tmatpow)

    return {
        'ck': ckeq,
        'time': times,
        'is_ergodic': is_ergodic,
        'is_fuzzy_ergodic': is_fuzzy_ergodic,
    }


def _chapman_kolmogorov_test_md(trajs, tmin, tmax, steps=30):
    r"""Calculate the Chapman Kolmogorov equation $T(n\tau)$."""
    times = np.around(np.geomspace(
        start=tmin,
        stop=tmax,
        num=steps,
    )).astype(np.int64)
    # filter duplicated times (for small ranges)
    times = np.unique(times)
    ntimes = len(times)

    ckeq = np.empty((trajs.nstates, ntimes))
    is_ergodic = np.empty(ntimes, dtype=bool)
    is_fuzzy_ergodic = np.empty(ntimes, dtype=bool)

    # enforce using non-lumped trajectory
    if isinstance(trajs, LumpedStateTraj):
        macrotrajs = StateTraj(trajs.state_trajs)
    else:
        macrotrajs = trajs

    # estimate Markov model
    for idx, time in enumerate(times):
        tmat, _ = macrotrajs.estimate_markov_model(lagtime=time)
        ckeq[:, idx] = np.diagonal(tmat)
        is_ergodic[idx] = tests.is_ergodic(tmat)
        is_fuzzy_ergodic[idx] = tests.is_fuzzy_ergodic(tmat)

    return {
        'ck': ckeq,
        'time': times,
        'is_ergodic': is_ergodic,
        'is_fuzzy_ergodic': is_fuzzy_ergodic,
    }


@decorit.alias('bh_test')
def buchete_hummer_test(trajs, lagtime, tmax):
    r"""Calculate the Buchete Hummer test.

    This method estimates the Buchete Hummer autocorrelation test. Projecting
    the state trajectory onto the right eigenvectors of the row normalized
    transition matrix
    $$C_{lm} (t) = \langle \phi_l[s(\tau +t)] \phi_m[S(\tau)]\rangle$$
    where \(\phi_i\) is the \(i\)-th right eigenvector. Buchete and Hummer[^2]
    showed that for a Markovian system it obeys an exponentil decay,
    corresponds to
    $$C_{lm} (t) = \delta_{lm} \exp(-t / t_k)$$
    with the implied timescale \(t_k = - \tau_\text{lag} / \ln \lambda_k\).

    [^2]: Buchete and Hummer
        **Coarse master equations for peptide folding dynamics**,
        *J. Phys. Chem.*, 112, 6057-6069 (2008),
        doi:[10.1021/jp0761665](https://doi.org/10.1021/jp0761665)

    Parameters
    ----------
    trajs : StateTraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    lagtime : int
        Lagtimes for estimating the markov model given in [frames].

    tmax : int
        Longest time to evaluate the CK equation given in [frames].

    Returns
    -------
    bhtest : dict
        Dictionary holding for each lagtime the ckequation and with 'md' the
        reference.

    """
    # format input
    trajs = StateTraj(trajs)

    # check that lag times are array of integers
    if not isinstance(lagtime, int) or lagtime < 0:
        raise TypeError(
            'Lagtimes needs to be positive integers, but {0}.'.format(lagtime),
        )

    if not isinstance(tmax, int) or tmax < 0:
        raise TypeError(
            'tmax needs to be a positive integer, but {0}'.format(tmax),
        )

    return _buchete_hummer_test(trajs, lagtime, tmax)


def _project_states_onto_vector(trajs, vector):
    """Shift states to corresponding vector values.

    Parameters
    ----------
    trajs : StateTraj or ndarray or list or list of ndarrays
        1D data or a list of data.

    vector : ndarray or list
        Values which will be used instead of states.

    Returns
    -------
    array : ndarray
        Shifted data in same shape as input.

    """
    trajs = StateTraj(trajs)
    vector = np.atleast_1d(vector)

    # check data-type
    if len(vector) != trajs.nstates:
        raise TypeError(
            'Vector of wrong length. {0} '.format(len(vector)) +
            'vs. {0}'.format(trajs.nstates),
        )

    # flatten data
    limits = np.cumsum([len(traj) for traj in trajs.trajs])
    array = trajs.trajs_flatten

    # convert to np.array
    states = np.arange(trajs.nstates)

    # shift data
    conv = np.arange(trajs.nstates, dtype=np.float64)
    conv[states] = vector
    array = conv[array]

    # reshape and return
    return np.split(array, limits)[:-1]


def _buchete_hummer_test(trajs, lagtime, tmax):
    r"""Calculate the Buchete Hummer equation."""
    times = _calc_times(lagtime=lagtime, tmax=tmax)
    ntimes = len(times)
    bheq = np.empty((trajs.nstates - 1, ntimes))

    # estimate Markov model
    tmat, _ = msm.estimate_markov_model(trajs, lagtime=lagtime)

    # symmetrize matrix
    tmat = np.sqrt(tmat * tmat.T)

    # get orthogonal eigenvectors of symmetric matrix
    evals, evecs = linalg.right_eigenvectors(tmat)
    pi, *evecs = evecs  # first eigenvector is equilibrium population

    # stationary distribution
    pi = pi**2 / np.sum(pi**2)

    # normalize eigenvectors to be eigenvectors of the non symmetric transition
    # matrix
    evecs = np.array(evecs) / np.sqrt(pi[np.newaxis, :])

    for idx, evec in enumerate(evecs):
        trajs_proj = _project_states_onto_vector(trajs, evec)

        if not numba.config.DISABLE_JIT:  # pragma: no cover
            trajs_proj = numba.typed.List(trajs_proj)

        bheq[idx] = _autocorrelation(trajs_proj, times)

        # for idx_col, time in enumerate(times):
        #     # autocorrelation function
        #     c_nn, norm = 0, 0
        #     for traj_proj in trajs_proj:
        #         c_nn += np.sum(traj_proj[: -time] * traj_proj[time:])
        #         norm += len(traj_proj[time:])
#
        #     bheq[idx, idx_col] = c_nn / norm

    # calculate reference (MD)
    bheq_ref = np.empty((trajs.nstates - 1, ntimes))
    for idx, eigenval in enumerate(evals[1:]):
        bheq_ref[idx] = np.exp(times * np.log(eigenval) / lagtime)

    return {
        lagtime: {'bh': bheq, 'time': times},
        'md': {'bh': bheq_ref, 'time': times},
    }


@numba.njit
def _autocorrelation(trajs, times):
    """Generate a simple transition count matrix from multiple trajectories."""
    # initialize matrix
    bheq = np.zeros(len(times), dtype=np.float64)

    for idx, time in enumerate(times):
        # autocorrelation function
        c_nn, norm = 0, 0
        for traj in trajs:
            for val_from, val_to in zip(  # noqa: WPS519
                traj[: -time], traj[time:],
            ):
                c_nn += val_from * val_to
            norm += len(traj[time:])

        bheq[idx] = c_nn / norm

    return bheq


def _calc_times(lagtime, tmax):
    """Return time array."""
    steps = int(np.floor(tmax / lagtime))
    return lagtime * np.arange(1, steps + 1)



# -*- coding: utf-8 -*-
"""Tests for the msm module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

Author: Daniel Nagel
        Georg Diez

"""
import numba
import numpy as np
import pytest

import msmhelper
from msmhelper import msm
from msmhelper.statetraj import StateTraj


@pytest.mark.parametrize('mat, matref', [
    ([[1, 1], [3, 1]], [[0.5, 0.5], [0.75, 0.25]])])
def test_row_normalize_matrix(mat, matref):
    """Test row normalization."""
    # cast mat to ndarray
    mat = np.array(mat)
    mat = msm.row_normalize_matrix(mat)
    np.testing.assert_array_equal(mat, matref)

    # set first row to 0
    mat[0] = 0
    np.testing.assert_array_equal(mat[0], msm.row_normalize_matrix(mat)[0])


@pytest.mark.parametrize('traj, lagtime, Tref, nstates', [
    ([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0], 1,
     [[0, 0, 1], [0, 4, 2], [2, 1, 2]], 3)])
def test__generate_transition_count_matrix(traj, lagtime, Tref, nstates):
    """Test transition count matrix estimate."""
    # convert traj to numba # noqa: SC100
    if numba.config.DISABLE_JIT:
        traj = [traj]
    else:
        traj = numba.typed.List([np.array(traj)])

    T = msm._generate_transition_count_matrix(traj, lagtime, nstates)
    np.testing.assert_array_equal(T, Tref)


@pytest.mark.parametrize('traj, lagtime, Tref, statesref', [
    ([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0], 1,
     [[0., 0., 1.], [0., 2 / 3, 1 / 3], [0.4, 0.2, 0.4]], [0, 1, 2]),
    ([3, 3, 3, 3, 3, 2, 2, 3, 2, 0, 2, 2, 0], 1,
     [[0., 1., 0.], [0.4, 0.4, 0.2], [0., 1 / 3, 2 / 3]], [0, 2, 3]),
])
def test_estimate_markov_model(traj, lagtime, Tref, statesref):
    """Test estimate markov model."""
    T, states = msmhelper.estimate_markov_model(traj, lagtime)
    np.testing.assert_array_equal(T, Tref)
    np.testing.assert_array_equal(states, statesref)


@pytest.mark.parametrize('traj, lagtime, Tref, statesref', [
    ([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0], 1,
     [[0., 0., 1.], [0., 2 / 3, 1 / 3], [0.4, 0.2, 0.4]], [0, 1, 2]),
])
def test__estimate_markov_model(traj, lagtime, Tref, statesref):
    """Test estimate markov model."""
    traj = StateTraj(traj)
    T, states = msm._estimate_markov_model(traj.trajs, lagtime, traj.nstates)
    np.testing.assert_array_equal(T, Tref)
    np.testing.assert_array_equal(states, statesref)


@pytest.mark.parametrize('transmat, lagtime, result', [
    (
        [[0.8, 0.2, 0.0], [0.2, 0.78, 0.02], [0.0, 0.2, 0.8]],
        2,
        -2 / np.log([4 / 5, 29 / 50]),
    ),
    (
        [[0.1, 0.9], [0.8, 0.2]],
        1,
        [np.nan],
    ),
])
def test__implied_timescales(transmat, lagtime, result):
    """Test implied timescale."""
    impl = msm._implied_timescales(transmat, lagtime)
    np.testing.assert_array_almost_equal(impl, result)


@pytest.mark.parametrize('trajs, lagtimes, result', [
    ([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1], [1, 2],
     [-1 / np.log([2 / 3]), -2 / np.log([7 / 15])])])
def test_implied_timescales(trajs, lagtimes, result):
    """Test implied timescale."""
    impl = msmhelper.implied_timescales(trajs, lagtimes)
    np.testing.assert_array_almost_equal(impl, result)

    with pytest.raises(TypeError):
        impl = msmhelper.implied_timescales(trajs, [-1, 2])

    with pytest.raises(TypeError):
        impl = msmhelper.implied_timescales(trajs, [1, 2.3])

    with pytest.raises(TypeError):
        impl = msmhelper.implied_timescales(trajs, lagtimes, reversible=True)


@pytest.mark.parametrize('tmat, peqref, kwargs, error', [
    (
        np.array([
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # nevver left
            [0.0, 0.8, 0.1, 0.1, 0.0, 0.0],
            [0.0, 0.1, 0.5, 0.4, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],  # nevver left
            [0.0, 0.2, 0.0, 0.0, 0.8, 0.0],  # never entered
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # never visited
        ]),
        [0, 0.6, 0.4, 0, 0, 0],
        {},
        None,
    ),
    (
        np.array([
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # nevver left
            [0.0, 0.8, 0.1, 0.1, 0.0, 0.0],
            [0.0, 0.1, 0.5, 0.4, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],  # nevver left
            [0.0, 0.2, 0.0, 0.0, 0.8, 0.0],  # never entered
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # never visited
        ]),
        None,
        {'allow_non_ergodic': False},
        ValueError,
    ),
    (
        np.array([[0.9, 0.1, 0.0], [0.1, 0.8, 0.1], [0.0, 0.1, 0.9]]),
        np.ones(3) / 3,
        {},
        None,
    ),
])
def test_equilibrium_population(tmat, peqref, kwargs, error):
    """Test peq."""
    if error is None:
        peq = msm.equilibrium_population(tmat, **kwargs)
        np.testing.assert_array_almost_equal(peq, peqref)
    else:
        with pytest.raises(error):
            msm.equilibrium_population(tmat, **kwargs)



# -*- coding: utf-8 -*-
"""Tests for the statetraj module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest
from numpy import array, int32

from msmhelper.statetraj import LumpedStateTraj, StateTraj


@pytest.fixture
def index_traj():
    """Define index trajectory."""
    traj = array(
        [0, 0, 0, 1, 1, 1, 2, 2, 2, 1, 1, 1, 0, 2, 1, 2, 2, 2],
        dtype=int32,
    )
    return StateTraj(traj)


@pytest.fixture
def indextraj():
    """Define index trajectory."""
    return array(
        [0, 0, 0, 1, 1, 1, 2, 2, 2, 1, 1, 1, 0, 2, 1, 2, 2, 2],
        dtype=int32,
    )


@pytest.fixture
def state_traj():
    """Define state trajectory."""
    traj = array(
        [0, 0, 0, 3, 3, 3, 2, 2, 2, 3, 3, 3, 0, 2, 3, 2, 2, 2],
        dtype=int32,
    )
    return StateTraj(traj)


@pytest.fixture
def statetraj():
    """Define state trajectory."""
    return array(
        [0, 0, 0, 3, 3, 3, 2, 2, 2, 3, 3, 3, 0, 2, 3, 2, 2, 2],
        dtype=int32,
    )


@pytest.fixture
def macrotraj():
    """Define state trajectory."""
    return array(
        [1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2],
        dtype=int32,
    )


@pytest.fixture
def macro_traj():
    """Define index trajectory."""
    traj = array(
        [1, 1, 1, 2, 2, 2, 3, 3, 3, 2, 2, 2, 1, 3, 2, 3, 3, 3],
        dtype=int32,
    )
    macrotraj = array(
        [1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2],
        dtype=int32,
    )
    return LumpedStateTraj(macrotrajs=macrotraj, microtrajs=traj)


def test_StateTraj_constructor(statetraj):
    """Test construction of object."""
    traj = StateTraj(statetraj)
    np.testing.assert_array_equal(
        statetraj,
        traj.state_trajs[0],
    )

    # check if passing StateTraj returns object
    assert traj is StateTraj(traj)
    # check if passing LumpedStateTraj returns object
    lumpedTraj = LumpedStateTraj(statetraj, statetraj)
    assert lumpedTraj is StateTraj(lumpedTraj)


def test_LumpedStateTraj_constructor(macrotraj, statetraj):
    """Test construction of object."""
    traj = LumpedStateTraj(macrotraj, statetraj)
    np.testing.assert_array_equal(
        macrotraj,
        traj.state_trajs[0],
    )

    np.testing.assert_array_equal(
        statetraj,
        traj.microstate_trajs[0],
    )

    # check if passing LumpedStateTraj returns object
    assert traj is LumpedStateTraj(traj)
    assert traj is StateTraj(traj)

    with pytest.raises(TypeError):
        LumpedStateTraj(macrotraj)


def test_nstates(state_traj, statetraj, macro_traj, macrotraj):
    """Test nstates property."""
    assert state_traj.nstates == len(np.unique(statetraj))
    assert macro_traj.nstates == len(np.unique(macrotraj))

    with pytest.raises(AttributeError):
        state_traj.nstates = 5


def test_nframes(state_traj):
    """Test nframes property."""
    assert state_traj.nframes == len(state_traj[0])

    with pytest.raises(AttributeError):
        state_traj.nframes = 5


def test_ntrajs(state_traj):
    """Test ntrajs property."""
    assert state_traj.ntrajs == len(state_traj.trajs)

    with pytest.raises(AttributeError):
        state_traj.ntrajs = 5


def test_index_trajs(state_traj, index_traj):
    """Test index trajs property."""
    np.testing.assert_array_equal(
        index_traj.trajs,
        index_traj.state_trajs,
    )
    np.testing.assert_array_equal(
        state_traj.trajs,
        state_traj.index_trajs,
    )

    with pytest.raises(AttributeError):
        state_traj.trajs = 5
    with pytest.raises(AttributeError):
        state_traj.index_trajs = 5


def test_trajs_flatten(state_traj, index_traj):
    """Test flatten index trajectory."""
    np.testing.assert_array_equal(
        index_traj.trajs[0],
        index_traj.trajs_flatten,
    )
    np.testing.assert_array_equal(
        state_traj.trajs[0],
        state_traj.trajs_flatten,
    )


def test_state_trajs_flatten(state_traj, index_traj):
    """Test flatten state trajectory."""
    np.testing.assert_array_equal(
        index_traj.state_trajs[0],
        index_traj.state_trajs_flatten,
    )
    np.testing.assert_array_equal(
        state_traj.state_trajs[0],
        state_traj.state_trajs_flatten,
    )


def test_microstate_trajs(macrotraj, statetraj, indextraj):
    """Test flatten trajectory."""
    macro_traj = LumpedStateTraj(macrotraj, statetraj)
    np.testing.assert_array_equal(
        statetraj,
        macro_traj.microstate_trajs[0],
    )
    np.testing.assert_array_equal(
        statetraj,
        macro_traj.microstate_trajs_flatten,
    )
    np.testing.assert_array_equal(
        macrotraj,
        macro_traj.state_trajs[0],
    )
    np.testing.assert_array_equal(
        macrotraj,
        macro_traj.state_trajs_flatten,
    )

    # check that state_trajs cannot be set unlike to StateTraj
    with pytest.raises(AttributeError):
        macro_traj.state_trajs = 5

    # check for index trajs
    macro_traj = LumpedStateTraj(macrotraj, indextraj)
    np.testing.assert_array_equal(
        indextraj,
        macro_traj.trajs[0],
    )
    np.testing.assert_array_equal(
        indextraj,
        macro_traj.trajs_flatten,
    )


def test___eq__(state_traj, index_traj):
    """Test eq method."""
    for traj in [state_traj, index_traj]:
        assert StateTraj(traj.state_trajs) == traj

    assert state_traj != index_traj
    assert state_traj != 5


def test_LumpedStateTraj__eq__(macro_traj):
    """Test eq method."""
    assert LumpedStateTraj(
        macro_traj.state_trajs, macro_traj.microstate_trajs,
    ) == macro_traj

    assert macro_traj != LumpedStateTraj([0, 0], [1, 0])
    assert macro_traj != 5


def test___repr__(state_traj, index_traj, macro_traj):
    """Test repr method."""
    for traj in [state_traj, index_traj, macro_traj]:
        assert eval(traj.__repr__()) == traj  # noqa: S307


def test___str__(state_traj, index_traj, macro_traj):
    """Test str method."""
    for traj in [state_traj, index_traj, macro_traj]:
        assert traj.__str__().startswith('[')


def test_as_list(state_traj, index_traj):
    """Test iterating over object."""
    for traj in [state_traj, index_traj]:
        for trajectory in traj:
            assert StateTraj(trajectory) == traj


def test_LumpedStateTraj_as_list(statetraj, indextraj, macrotraj):
    """Test iterating over object."""
    for traj in [statetraj, indextraj, macrotraj]:
        macro_traj = LumpedStateTraj(traj, traj)
        for trajectory in macro_traj:
            assert LumpedStateTraj(trajectory, trajectory) == macro_traj

        np.testing.assert_array_equal(macro_traj[0], traj)


def test_LumpedStateTraj_estimate_markov_model(macro_traj):
    """Check MSM estimation from microstate matrix."""
    tlag = 1
    states_ref = np.array([1, 2])
    tmat_ref = np.array([[0.50177725, 0.49822275], [0.06872038, 0.93127962]])

    tmat, states = macro_traj.estimate_markov_model(tlag)
    np.testing.assert_array_almost_equal(tmat, tmat_ref)
    np.testing.assert_array_equal(states, states_ref)

    with pytest.raises(TypeError):
        macrotraj = macro_traj.state_trajs_flatten
        microtraj = macro_traj.microstate_trajs_flatten
        microtraj[-1] = np.max(microtraj) + 1
        trap_traj = LumpedStateTraj(macrotraj, microtraj)
        trap_traj.estimate_markov_model(tlag)


@pytest.mark.parametrize('traj, state, idx, error', [
    ([1, 2, 4], 1, 0, None),
    ([1, 2, 4], 0, None, ValueError),
    ([1, 8, 3], 8, 2, None),
    ([1, 8, 3, 1], 3, 1, None),
])
def test_state_to_idx(traj, state, idx, error):
    """Check state to idx conversion."""
    statetraj = StateTraj(traj)
    for straj in (
        statetraj,
        LumpedStateTraj(statetraj, np.arange(statetraj.nframes)),
    ):
        if error is not None:
            with pytest.raises(error):
                straj.state_to_idx(state)
        else:
            assert straj.state_to_idx(state) == idx



# -*- coding: utf-8 -*-
"""Tests for the tests module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest

from msmhelper import tests, tools
from msmhelper.statetraj import StateTraj


@pytest.mark.parametrize('mat', [(np.arange(9).reshape(3, 3))])
def test_is_quadratic(mat):
    """Test is_quadratic."""
    # check if no error is raised
    assert tests.is_quadratic(mat)
    # check for non quadratic matrices
    assert not tests.is_quadratic([2, 1])
    # check for scalar
    assert not tests.is_quadratic(1)
    # check for 3d matrices
    assert not tests.is_quadratic(np.arange(8).reshape(2, 2, 2))


@pytest.mark.parametrize('traj', [([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0])])
def test_is_state_traj(traj):
    """Test formating state trajectory."""
    traj = tools.format_state_traj(traj)

    assert tests.is_state_traj(traj)
    assert not tests.is_state_traj([traj[0].astype(np.float32)])


@pytest.mark.parametrize('traj', [([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0])])
def test_is_index_traj(traj):
    """Test formating state trajectory."""
    traj = tools.format_state_traj(traj)

    assert tests.is_index_traj(traj)
    assert tests.is_index_traj(StateTraj(traj))
    traj.append(np.array([4, 5]))
    assert not tests.is_index_traj(traj)
    assert not tests.is_index_traj([5])


def test_is_tmat():
    """Test is_tmat."""
    mat = np.array([[0.9, 0.1], [0.2, 0.8]])

    # check if no error is raised
    assert tests.is_tmat(mat)

    # check if row and cols of zero is allowed
    mat2 = np.zeros((3, 3))
    mat2[:2, :2] = mat
    assert tests.is_tmat(mat2)

    # check if only row fails
    mat2[0, 1:] = mat2[0, 2], mat2[0, 1]
    assert not tests.is_tmat(mat2)

    mat[0, 0] = 0.8
    # check for non normalized matrices
    assert not tests.is_tmat(mat)

    # check for non quadratic matrices
    assert not tests.is_quadratic(mat[0])


def test_is_ergodic():
    """Test is_ergodic."""
    # if not ergodic
    mat = np.array([[0.9, 0.1, 0.0], [0.1, 0.9, 0.0], [0.2, 0.2, 0.6]])
    assert not tests.is_ergodic(mat)

    # if ergodic
    mat = np.array([[0.9, 0.1, 0.0], [0.1, 0.8, 0.1], [0.2, 0.2, 0.6]])
    assert tests.is_ergodic(mat)

    # if is not a transition matrix
    mat[0, 0] = 0.8
    assert not tests.is_ergodic(mat)


def test_is_fuzzy_ergodic():
    """Test is_fuzzy_ergodic."""
    # if not ergodic
    mat = np.array([[0.9, 0.1, 0.0], [0.1, 0.9, 0.0], [0.2, 0.2, 0.6]])
    assert not tests.is_fuzzy_ergodic(mat)

    # if ergodic
    mat = np.array([[0.9, 0.1, 0.0], [0.1, 0.8, 0.1], [0.2, 0.2, 0.6]])
    assert tests.is_fuzzy_ergodic(mat)

    # if ergodic with trap state
    mat = np.array([
        [0.9, 0.1, 0.0, 0.0],
        [0.1, 0.8, 0.1, 0.0],
        [0.2, 0.2, 0.6, 0.0],
        [0.0, 0.0, 0.0, 1.0]
    ])
    assert tests.is_fuzzy_ergodic(mat)

    # if ergodic with not visited state
    mat = np.array([
        [0.9, 0.1, 0.0, 0.0],
        [0.1, 0.8, 0.1, 0.0],
        [0.2, 0.2, 0.6, 0.0],
        [0.0, 0.0, 0.0, 0.0]
    ])
    assert tests.is_fuzzy_ergodic(mat)

    # if is not a transition matrix
    mat[0, 0] = 0.8
    assert not tests.is_fuzzy_ergodic(mat)


def test_ergodic_mask():
    """Test ergodic_mask."""
    # if ergodic
    mat = np.array([
        [0.9, 0.1, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.8, 0.1, 0.1, 0.0, 0.0],
        [0.0, 0.1, 0.8, 0.1, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.8, 0.1, 0.1],
        [0.0, 0.2, 0.0, 0.0, 0.8, 0.0],
        [0.1, 0.0, 0.0, 0.0, 0.0, 0.9],
    ])
    mask = np.array([True, True, True, True, True, True])
    np.testing.assert_array_equal(tests.ergodic_mask(mat), mask)

    mat = np.array([
        [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # nevver left
        [0.0, 0.8, 0.1, 0.1, 0.0, 0.0],
        [0.0, 0.1, 0.8, 0.1, 0.0, 0.0],
        [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],  # nevver left
        [0.0, 0.2, 0.0, 0.0, 0.8, 0.0],  # never entered
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # never visited
    ])
    mask = np.array([False, True, True, False, False, False])
    np.testing.assert_array_equal(tests.ergodic_mask(mat), mask)

    # if is not a transition matrix
    mat[0, 0] = 0.8

    with pytest.raises(ValueError):
        tests.ergodic_mask(mat)



# -*- coding: utf-8 -*-
"""Set of functions for analyzing the MD trajectory.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import random
from collections import defaultdict

import decorit
import numba
import numpy as np

from msmhelper.compare import _intersect as intersect
from msmhelper.statetraj import StateTraj


@decorit.alias('estimate_wt')
def estimate_waiting_times(trajs, start, final):
    """Estimates waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from first entering the start-basin until first
    reaching the final-basin.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.

    Returns
    -------
    wt : ndarray
        List of waiting times, given in frames.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectoty.',
            )

    # do not convert for pytest coverage
    if numba.config.DISABLE_JIT:
        return _estimate_waiting_times(trajs, states_start, states_final)
    return _estimate_waiting_times(  # pragma: no cover
        numba.typed.List(trajs),
        numba.typed.List(states_start),
        numba.typed.List(states_final),
    )


@numba.njit
def _estimate_waiting_times(trajs, states_start, states_final):
    """Estimate waiting time between subsets of states."""
    times = []
    for traj in trajs:
        times.extend(
            _estimate_waiting_times_singletraj(
                traj, states_start, states_final,
            ),
        )
    return np.array(times)


@numba.njit
def _estimate_waiting_times_singletraj(traj, states_start, states_final):
    """Estimate waiting time between subsets of states for a single traj."""
    paths_idx = _estimate_events_singletraj(traj, states_start, states_final)
    return [idx_end - idx_start for idx_start, idx_end in paths_idx]


def estimate_paths(trajs, start, final):
    """Estimates paths and waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from first entering the start-basin until first
    reaching the final-basin. The results will be listed by the corresponding
    pathways, where loops are removed occuring first.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.

    Returns
    -------
    paths : ndarray
        List of waiting times, given in frames.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectoty.',
            )

    # do not convert for pytest coverage
    if numba.config.DISABLE_JIT:  # pragma: no cover
        path_tuples = _estimate_paths(trajs, states_start, states_final)
    else:
        path_tuples = _estimate_paths(  # pragma: no cover
            numba.typed.List(trajs),
            numba.typed.List(states_start),
            numba.typed.List(states_final),
        )
    paths = defaultdict(list)
    for path, time in path_tuples:
        paths[tuple(path)].append(time)

    return paths


@numba.njit
def _estimate_paths(trajs, states_start, states_final):
    """Estimate waiting time between subsets of states."""
    paths = []
    for traj in trajs:
        paths.extend(
            _estimate_paths_singletraj(
                traj, states_start, states_final,
            ),
        )
    return paths


@numba.njit
def _estimate_events_singletraj(traj, states_start, states_final):
    """Estimate waiting time between subsets of states for a single traj."""
    paths_idx = []
    propagates_forwards = False
    idx_start = 0
    for idx in range(len(traj)):  # noqa: WPS518
        state = traj[idx]

        if not propagates_forwards and state in states_start:
            propagates_forwards = True
            idx_start = idx
        elif propagates_forwards and state in states_final:
            propagates_forwards = False
            paths_idx.append((idx_start, idx))

    return paths_idx


@numba.njit
def _estimate_paths_singletraj(traj, states_start, states_final):
    """Estimate waiting time between subsets of states for a single traj."""
    paths_idx = _estimate_events_singletraj(traj, states_start, states_final)

    paths = []
    for idx_start, idx_end in paths_idx:
        for state in traj[idx_start: idx_end + 1]:  # noqa: WPS518
            if state in states_start:
                path = [int(define_type_var) for define_type_var in range(0)]
            elif state in path:
                path = path[: path.index(state)]
            path.append(state)

        paths.append((path, idx_end - idx_start))
    return paths


@decorit.alias('estimate_msm_wt')
def estimate_msm_waiting_times(
    *,
    trajs,
    lagtime,
    start,
    final,
    steps,
    return_list=False,
):
    """Estimates waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from first entering the start-basin until first
    reaching the final-basin.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    lagtime : int
        Lag time for estimating the markov model given in [frames].

    start : int or list of
        States to start counting.

    final : int or list of
        States to start counting.

    steps : int
        Number of MCMC propagation steps of MCMC run.

    return_list : bool
        If true a list of all events is returned, else a dictionary is
        returned.

    Returns
    -------
    wt : ndarray
        List of waiting times, given in frames.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectoty.',
            )

    # convert states to idx
    idxs_start = np.array(
        [trajs.state_to_idx(state) for state in states_start],
    )
    idxs_final = np.array(
        [trajs.state_to_idx(state) for state in states_final],
    )
    start = np.random.choice(idxs_final)

    # do not convert for pytest coverage
    if not numba.config.DISABLE_JIT:
        idxs_start = numba.typed.List(idxs_start)
        idxs_final = numba.typed.List(idxs_final)

    # estimate cummulative transition matrix
    cummat = _get_cummat(trajs=trajs, lagtime=lagtime)

    wts = _estimate_msm_waiting_times(
        cummat=cummat,
        start=start,
        states_from=idxs_start,
        states_to=idxs_final,
        steps=steps,
    )
    # multiply wts by lagtime
    if return_list:
        return np.repeat(
            list(wts.keys()), list(wts.values()),
        ) * lagtime
    return {wt * lagtime: count for wt, count in wts.items()}


@numba.njit
def _propagate_MCMC_step(cummat, idx_from):
    """Propagate a single step Markov chain Monte Carlo."""
    rand = random.random()  # noqa: S311
    cummat_perm, state_perm = cummat
    cummat_perm, state_perm = cummat_perm[idx_from], state_perm[idx_from]

    for idx, cummat_idx in enumerate(cummat_perm):
        # strict less to ensure that rand=0 does not jump along unconnected
        # states with Tij=0.
        if rand < cummat_idx:
            return state_perm[idx]
    # this should never be reached, but needed for numba to ensure int return
    return len(cummat_perm) - 1


@numba.njit
def _estimate_msm_waiting_times(
    cummat, start, states_from, states_to, steps,
):
    wts = {}

    idx_start = 0
    propagates_forwards = False
    state = start
    for idx in range(steps):
        state = _propagate_MCMC_step(cummat=cummat, idx_from=state)

        if not propagates_forwards and state in states_from:
            propagates_forwards = True
            idx_start = idx
        elif propagates_forwards and state in states_to:
            propagates_forwards = False
            wt = idx - idx_start
            if wt in wts:
                wts[wt] += 1
            else:
                wts[wt] = 1

    return wts


def propagate_MCMC(
    trajs,
    lagtime,
    steps,
    start=-1,
):
    """Propagate MCMC trajectory.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.

    lagtime : int
        Lag time for estimating the markov model given in [frames].

    steps : int
        Number of MCMC propagation steps.

    start : int or list of, optional
        State to start propagating. Default (-1) is random state.

    Returns
    -------
    mcmc : ndarray
        MCMC trajecory.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    # check that all states exist in trajectory
    if start == -1:
        start = np.random.choice(trajs.states)
    elif start not in trajs.states:
        raise ValueError(
            'Selected starting state does not exist in state trajectoty.',
        )

    # convert states to idx
    idx_start = trajs.state_to_idx(start)

    # estimate permuted cummulative transition matrix
    cummat = _get_cummat(trajs=trajs, lagtime=lagtime)

    # do not convert for pytest coverage
    return _propagate_MCMC(  # pragma: no cover
        cummat=cummat,
        start=idx_start,
        steps=steps,
    )


@numba.njit
def _propagate_MCMC(cummat, start, steps):
    mcmc = np.empty(steps, dtype=np.int32)

    state = start
    mcmc[0] = state
    for idx in range(steps - 1):
        state = _propagate_MCMC_step(cummat=cummat, idx_from=state)
        mcmc[idx + 1] = state

    return mcmc


def _get_cummat(trajs, lagtime):
    # estimate cummulative transition matrix
    msm, _ = StateTraj(trajs).estimate_markov_model(lagtime)

    cummat_perm = np.empty_like(msm)
    state_perm = np.empty_like(msm, dtype=np.int64)

    for idx, row in enumerate(msm):
        idx_sort = np.argsort(row)[::-1]

        cummat_perm[idx] = np.cumsum(row[idx_sort])
        state_perm[idx] = idx_sort

    cummat_perm[:, -1] = 1  # enforce that probability sums up to 1
    return cummat_perm, state_perm



# -*- coding: utf-8 -*-
"""Input and output text files.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import sys

import numpy as np
import pandas as pd

from msmhelper import tools


class FileError(Exception):
    """An exception for wrongly formated input files."""


def opentxt(file_name, comment='#', nrows=None, **kwargs):
    r"""Open a text file.

    This method can load an nxm array of floats from an ascii file. It uses
    either pandas read_csv for a single comment or as fallback the slower numpy
    laodtxt for multiple comments.

    .. warning:: In contrast to pandas the order of usecols will be used. So if
        using ´data = opentxt(..., uscols=[1, 0])´ you acces the first column
        by `data[:, 0]` and the second one by `data[:, 1]`.

    Parameters
    ----------
    file_name : string
        Name of file to be opened.

    comment : str or array of str, optional
        Characters with which a comment starts.

    nrows : int, optional
        The maximum number of lines to be read

    usecols : int-array, optional
        Columns to be read from the file (zero indexed).

    skiprows : int, optional
        The number of leading rows which will be skipped.

    dtype : data-type, optional
        Data-type of the resulting array. Default: float.

    Returns
    -------
    array : ndarray
        Data read from the text file.

    """
    if len(comment) == 1:
        # pandas does not support array of single char
        if not isinstance(comment, str):
            comment = comment[0]

        # force pandas to load in stated order without sorting
        cols = kwargs.pop('usecols', None)
        if cols is not None:
            idx = np.argsort(cols)
            cols = np.atleast_1d(cols).astype(np.int32)[idx]

        array = pd.read_csv(
            file_name,
            sep=r'\s+',
            header=None,
            comment=comment,
            nrows=nrows,
            usecols=cols,
            **kwargs,
        ).values

        if array.shape[-1] == 1:
            array = array.flatten()
        # swap columns back to ensure correct order
        elif cols is not None:
            array = tools.swapcols(array, idx, np.arange(len(idx)))

        return array

    return np.loadtxt(
        file_name,
        comments=comment,
        max_rows=nrows,
        **kwargs,
    )


def savetxt(file_name, array, header=None, fmt='%.5f'):  # noqa: WPS323
    """Save nxm array of floats to a text file.

    It uses numpys savetxt method and extends the header with information
    of execution.

    Parameters
    ----------
    file_name : string
        File name to store data.

    array : ndarray
        Data to be stored.

    header : str, optional
        Comment written into the header of the output file.

    fmt : str or sequence of strs, optional
        See numpy.savetxt fmt.

    """
    # prepare header comments
    RUI = tools.get_runtime_user_information()

    header_comment = (
        'This file was generated by {script_dir}/{script_name}:\n{args}' +
        '\n\n{date}, {user}@{pc}'
    ).format(**RUI, args=' '.join(sys.argv))

    if header:  # print column title if given
        header_comment += '\n{0}'.format(header)

    # save file
    np.savetxt(file_name, array, fmt=fmt, header=header_comment)


def opentxt_limits(file_name, limits_file=None, **kwargs):
    """Load file and split according to limit file.

    If limits_file is not provided it will return [traj].

    Parameters
    ----------
    file_name : string
        Name of file to be opened.

    limits_file : str, optional
        File name of limit file. Should be single column ascii file.

    kwargs
        The Parameters defined in opentxt.

    Returns
    -------
    traj : ndarray
        Return array of subtrajectories.

    """
    # open trajectory
    traj = opentxt(file_name, **kwargs)

    # open limits
    limits = open_limits(limits_file=limits_file, data_length=len(traj))

    # split trajectory
    return np.split(traj, limits)[:-1]


def openmicrostates(file_name, limits_file=None, **kwargs):
    """Load 1d file and split according to limit file.

    Both, the limit file and the trajectory file needs to be a single column
    file. If limits_file is not provided it will return [traj]. The trajectory
    will of dtype np.int16, so the states needs to be smaller than 32767.

    Parameters
    ----------
    file_name : string
        Name of file to be opened.

    limits_file : str, optional
        File name of limit file. Should be single column ascii file.

    kwargs
        The Parameters defined in opentxt.

    Returns
    -------
    traj : ndarray
        Return array of subtrajectories.

    """
    # open trajectory
    if 'dtype' in kwargs and not np.issubdtype(kwargs['dtype'], np.integer):
        raise TypeError('dtype should be integer')
    else:
        kwargs['dtype'] = np.int16

    # load split trajectory
    traj = opentxt_limits(file_name, limits_file, **kwargs)

    if len(traj[0].shape) != 1:
        raise FileError('Microstate trjectory shoud be single column file.')

    return traj


def open_limits(data_length, limits_file=None):
    """Load and check limit file.

    The limits give the length of each single trajectory. So e.g.
    [5, 5, 5] for 3 equally-sized subtrajectories of length 5.

    Parameters
    ----------
    data_length : int
        Length of data read.

    limits_file : str, optional
        File name of limit file. Should be single column ascii file.

    Returns
    -------
    limits : ndarray
        Return cumsum of limits.

    """
    if limits_file is None:
        return np.array([data_length])  # for single trajectory

    # open limits file
    limits = opentxt(limits_file)
    if len(limits.shape) != 1:
        raise FileError('Shoud be single column file.')

    # convert to cumulative sum
    limits = np.cumsum(limits)
    if data_length != limits[-1]:
        raise ValueError('Limits are inconsistent to data.')

    return limits



# -*- coding: utf-8 -*-
import pathlib
import sys
from collections import defaultdict

import setuptools

# check for python version
if sys.version_info < (3, 6):
    raise SystemExit('Python 3.6+ is required!')


def get_extra_requirements(path, add_all=True):
    """Parse extra-requirements file."""
    with open(path) as depfile:
        extra_deps = defaultdict(set)
        for line in depfile:
            if not line.startswith('#'):
                if ':' not in line:
                    raise ValueError(
                        f'Dependency in {path} not correct formatted: {line}',
                    )
                dep, tags = line.split(':')
                tags = {tag.strip() for tag in tags.split(',')}
                for tag in tags:
                    extra_deps[tag].add(dep)

        # add tag `all` at the end
        if add_all:
            extra_deps['all'] = {
                tag for tags in extra_deps.values() for tag in tags
            }

    return extra_deps


def remove_gh_dark_mode_only_tags(text, tag='#gh-dark-mode-only'):
    """Remove recursively all """
    idx = text.find(tag)
    if idx < 0:
        return text

    idx_tag_end = text.find('>', idx)
    idx_tag_start = text.rfind('<', 0, idx)
    return remove_gh_dark_mode_only_tags(
        text[:idx_tag_start] + text[idx_tag_end + 1:], tag,
    )


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = remove_gh_dark_mode_only_tags(
    (HERE / 'README.md').read_text(),
)

# This call to setup() does all the work
setuptools.setup(
    name='msmhelper',
    version='0.6.1',
    description='Helper functions for Markov State Models.',
    long_description=README,
    long_description_content_type='text/markdown',
    keywords=[
        'MSM',
        'Markov model',
        'Markov state model',
        'MD analysis',
    ],
    author='braniii',
    url='https://github.com/moldyn/msmhelper',
    license='BSD-3-Clause License',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={
        'Documentation': 'https://moldyn.github.io/msmhelper',
        'Source Code': 'https://github.com/moldyn/msmhelper',
        'Bug Tracker': 'https://github.com/moldyn/msmhelper/issues',
    },
    packages=setuptools.find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
    install_requires=[
        'numpy',
        'numba',
        'pandas',
        'decorit',
        'scipy',
    ],
    extras_require=get_extra_requirements('extra-requirements.txt'),
)



# -*- coding: utf-8 -*-
"""Tests for the tools module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
from importlib import reload

import numpy as np
import pytest

import __main__ as main
import msmhelper
from msmhelper import tools
from msmhelper.statetraj import StateTraj


class change_main__file__:
    """Emulate an ipython usage."""

    def __enter__(self):
        """Delete file name and reload module."""
        self.filename = main.__file__
        del main.__file__
        main.msmhelper = reload(msmhelper)

    def __exit__(self, typ, val, traceback):
        """Reset file name and reload module."""
        main.__file__ = self.filename
        main.msmhelper = reload(msmhelper)


@pytest.mark.parametrize('data, expected, val_old, val_new', [
    ([1, 1, 1, 3, 2, 2], [2, 2, 2, 3, 1, 1], [1, 2, 3], [2, 1, 3]),
    ([1, -1, 1, 3, 2, 2], [2, -1, 2, 3, 1, 1], [1, 2], [2, 1]),
    ([1, 1, 1, 3, 2, 2], [3, 3, 3, 2, 1, 1], [1, 2, 3], [3, 1, 2]),
    (np.array([1, 1, 1, 3, 2, 2]), [3, 3, 3, 2, 1, 1], [1, 2, 3], [3, 1, 2]),
    ([np.array([1, 1, -5, 3, 2])], [[3, 3, -5, 2, 1]], [1, 2, 3], [3, 1, 2]),
    ([[1, 1, 1], [3, 2, 2]], [[2, 2, 2], [3, 1, 1]], [1, 2, 3], [2, 1, 3])])
def test_shift_data(data, expected, val_old, val_new):
    """Test shift_data."""
    data_shifted = msmhelper.shift_data(data, val_old, val_new)
    for i, row in enumerate(data_shifted):
        assert (row == expected[i]).all()

    with pytest.raises(TypeError):
        msmhelper.shift_data(data, val_old, val_new, dtype=np.float64)


@pytest.mark.parametrize('data, expected, perm_expected', [
    ([1, 3, 3, 3, 2, 2], [3, 1, 1, 1, 2, 2], [3, 2, 1]),
    ([1, -5, -5, 7, 7, 7], [3, 2, 2, 1, 1, 1], [7, -5, 1])])
def test_rename_by_population(data, expected, perm_expected):
    """Test rename_by_population."""
    assert (msmhelper.rename_by_population(data) == expected).all()
    result, perm = msmhelper.rename_by_population(
        data, return_permutation=True,
    )
    assert (result == expected).all()
    assert (perm == perm_expected).all()


@pytest.mark.parametrize('data, expected, perm_expected', [
    ([1, 3, 3, 3, 2, 2], [0, 2, 2, 2, 1, 1], [1, 2, 3]),
    ([1, -5, -5, 7, 7, 7], [1, 0, 0, 2, 2, 2], [-5, 1, 7])])
def test_rename_by_index(data, expected, perm_expected):
    """Test rename_by_index."""
    assert (msmhelper.rename_by_index(data) == expected).all()
    result, perm = msmhelper.rename_by_index(
        data, return_permutation=True,
    )
    assert (result == expected).all()
    assert (perm == perm_expected).all()


@pytest.mark.parametrize('data, expected, pop', [
    ([1, 3, 3, 3, 2, 2], [1, 2, 3], [1, 2, 3]),
    ([[1, -5, -5, 7, 7, 7], [1, 1]], [-5, 1, 7], [2, 3, 3])])
def test_unique(data, expected, pop):
    """Test unique."""
    assert (msmhelper.unique(data) == expected).all()
    result, perm = msmhelper.unique(data, return_counts=True)
    assert (result == expected).all()
    assert (perm == pop).all()


@pytest.mark.parametrize('data, expected, window', [
    ([1, 1, 1, 2, 2, 2, 3, 3, 3], [1., 1., 1., 2., 2., 2., 3., 3., 3.], 1),
    ([1, 1, 1, 2, 2, 2, 3, 3, 3], [.5, 1., 1., 1.5, 2., 2., 2.5, 3., 3.], 2)])
def test_runningmean(data, expected, window):
    """Test runningmean."""
    assert (msmhelper.runningmean(data, window) == expected).all()


@pytest.mark.parametrize('traj', [([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0])])
def test_format_state_traj(traj):
    """Test formating state trajectory."""
    # as list of floats
    with pytest.raises(TypeError):
        tools.format_state_traj([float(s) for s in traj])

    # as list of integers
    assert (tools.format_state_traj(traj)[0] == traj).all()

    # as list of lists
    assert (tools.format_state_traj([traj])[0] == traj).all()

    # as ndarray
    traj = np.array(traj)
    assert (tools.format_state_traj(traj)[0] == traj).all()

    # as 2d ndarray
    traj = np.atleast_2d(traj)
    assert (tools.format_state_traj(traj)[0] == traj).all()

    # as ndarray of floats
    with pytest.raises(TypeError):
        tools.format_state_traj(traj.astype(np.float64))

    # as list of ndarrays
    assert (tools.format_state_traj([traj])[0] == traj).all()


@pytest.mark.parametrize('data', [
    ([[1, 1], [1, 2], [2, 1], [2, 3], [3, 3]])])
def test_swapcols(data):
    """Test swapcols."""
    # for same indices
    data_swap = tools.swapcols(data, (0, 1), (0, 1))
    for row in range(len(data)):
        assert (data[row] == data_swap[row]).all()

    data_swap = tools.swapcols(data, (0, 1), (1, 0))
    for row in range(len(data)):
        assert (data[row][0] == data_swap[row][1]).all()
        assert (data[row][1] == data_swap[row][0]).all()

    # wrong shape
    with pytest.raises(ValueError):
        tools.swapcols(data, (0, 1), (1))


def test__asindex():
    """Test asindex."""
    idx = np.arange(5)
    assert (idx == tools._asindex(idx)).all()

    # check if integer is casted to array
    assert (np.array(5) == tools._asindex(5)).all()

    # wrong dimensionality
    with pytest.raises(ValueError):
        tools._asindex([idx])


@pytest.mark.parametrize('traj', [([1, 1, 1, 1, 1, 2, 2, 1, 2, 0, 2, 2, 0])])
def test__check_state_traj(traj):
    """Test if state trajectory is formatted."""
    traj = tools.format_state_traj(traj)

    assert tools._check_state_traj(traj)
    assert tools._check_state_traj(StateTraj(traj))

    with pytest.raises(TypeError):
        tools._check_state_traj([traj[0].astype(np.float32)])

    with pytest.raises(TypeError):
        tools._check_state_traj(traj[0])


def test_get_runtime_user_information():
    """Check for console usage."""
    assert tools.get_runtime_user_information()['script_name'] != 'console'
    with change_main__file__():
        assert tools.get_runtime_user_information()['script_name'] == 'console'


@pytest.mark.parametrize('mat, power', [
    (np.arange(16, dtype=np.float64).reshape(4, 4), 2),
])
def test_matrix_power(mat, power):
    """Test matrix power."""
    np.testing.assert_array_equal(
        tools.matrix_power(mat, power),
        np.linalg.matrix_power(mat, power),
    )


@pytest.mark.parametrize('array, val, pos', [
    (np.arange(10), 3, 3), (np.arange(10), 10, -1),
])
def test_find_first(array, val, pos):
    """Test find_first."""
    assert (tools.find_first(val, array) == pos)



# -*- coding: utf-8 -*-
"""Tests for the linalg module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
# ~~~ IMPORT ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
import numpy as np
import pytest

from msmhelper import linalg


# ~~~ TESTS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@pytest.mark.parametrize('matrix, eigenvaluesref, eigenvectorsref', [
    (np.array([[1, 6, -1], [2, -1, -2], [1, 0, -1]]), np.array([3, 0, -4]),
     [np.array([-2, -3, 2]) / np.sqrt(17),
      np.array([-1, -6, 13]) / np.sqrt(206),
      np.array([-1, 2, 1]) / np.sqrt(6)])])
def test_left_eigenvectors(matrix, eigenvaluesref, eigenvectorsref):
    """Test left eigenvectors estimate."""
    eigenvalues, eigenvectors = linalg.left_eigenvectors(matrix)
    np.testing.assert_array_almost_equal(eigenvalues, eigenvaluesref)
    np.testing.assert_array_almost_equal(
        np.abs(eigenvectors),
        np.abs(eigenvectorsref),
    )

    with pytest.raises(TypeError):
        linalg.left_eigenvectors(matrix[0])

    # test eigenvalues method
    eigenvalues = linalg.left_eigenvalues(matrix)
    np.testing.assert_array_almost_equal(eigenvalues, eigenvaluesref)

    with pytest.raises(TypeError):
        linalg.left_eigenvalues(matrix[0])


@pytest.mark.parametrize('matrix, eigenvaluesref, eigenvectorsref', [
    (np.array([[1, 6, -1], [2, -1, -2], [1, 0, -1]]), np.array([3, 0, -4]),
     [np.array([8, 3, 2]) / np.sqrt(77),
      np.array([1, 0, 1]) / np.sqrt(2),
      np.array([-9, 8, 3]) / np.sqrt(154)])])
def test_right_eigenvectors(matrix, eigenvaluesref, eigenvectorsref):
    """Test left eigenvectors estimate."""
    eigenvalues, eigenvectors = linalg.right_eigenvectors(matrix)
    np.testing.assert_array_almost_equal(eigenvalues, eigenvaluesref)
    np.testing.assert_array_almost_equal(
        np.abs(eigenvectors),
        np.abs(eigenvectorsref),
    )

    with pytest.raises(TypeError):
        linalg.right_eigenvectors(matrix[0])

    # test eigenvalues method
    eigenvalues = linalg.right_eigenvalues(matrix)
    np.testing.assert_array_almost_equal(eigenvalues, eigenvaluesref)

    with pytest.raises(TypeError):
        linalg.right_eigenvalues(matrix[0])



# -*- coding: utf-8 -*-
"""Tests for the iotext module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import os.path

import numpy as np
import pytest

from msmhelper import iotext

# Current directory
HERE = os.path.dirname(__file__)


@pytest.fixture
def limits_file():
    """Define limits file."""
    return os.path.join(HERE, 'limits.dat')


@pytest.mark.parametrize('traj_file, kwargs', [
    (os.path.join(HERE, 'data.dat'), {'comment': '#'}),
    (os.path.join(HERE, 'data.dat'), {'usecols': (1, 0)}),
    (os.path.join(HERE, 'traj1.dat'), {}),
    (os.path.join(HERE, 'traj1.dat'), {'comment': ['#']}),
    (os.path.join(HERE, 'traj2.dat'), {'comment': ['#', '@']}),
])
def test_opentxt(traj_file, kwargs):
    """Test that a file is opened correctly."""
    expected = np.array([[1, 2], [1, 3], [1, 2], [1, 1], [1, 2], [2, 1],
                         [2, 5], [1, 4], [2, 3], [3, 2], [2, 1], [2, 2],
                         [3, 3]])
    data = iotext.opentxt(traj_file, **kwargs)
    if len(data.shape) == 1:
        assert (data == expected[:, 0]).all()
    else:
        if 'usecols' in kwargs:
            expected[:, (0, 1)] = expected[:, kwargs['usecols']]
        for i, row in enumerate(data):
            assert (row == expected[i]).all()


@pytest.mark.parametrize('traj_file, header', [
    (os.path.join(HERE, 'traj1.dat'), None),
    (os.path.join(HERE, 'traj1.dat'), 'Test header comment.'),
])
def test_savetxt(traj_file, header, tmpdir):
    """Test that a file is opened correctly."""
    expected = [1, 1, 1, 1, 1, 2, 2, 1, 2, 3, 2, 2, 3]
    output = tmpdir.join('test_traj.txt')
    iotext.savetxt(output, expected, header=header, fmt='%.0f')
    assert (iotext.opentxt(output) == expected).all()


@pytest.mark.parametrize('wrong_limit', [os.path.join(HERE, 'data.dat')])
def test_open_limits(limits_file, wrong_limit):
    """Test that the limits are loaded correctly."""
    assert (iotext.open_limits(13, limits_file) == [5, 10, 13]).all()
    assert (iotext.open_limits(13) == [13]).all()
    with pytest.raises(iotext.FileError):
        iotext.open_limits(13, wrong_limit)
    with pytest.raises(ValueError):
        iotext.open_limits(14, limits_file)


@pytest.mark.parametrize('traj_file, kwargs', [
    (os.path.join(HERE, 'traj1.dat'), {}),
    (os.path.join(HERE, 'traj1.dat'), {'dtype': np.int32}),
    (os.path.join(HERE, 'traj1.dat'), {'dtype': np.float64}),
    (os.path.join(HERE, 'data.dat'), {'no_traj': True}),
])
def test_openmicrostates(limits_file, traj_file, kwargs):
    """Test that the trajectory is split correctly."""
    if 'dtype' in kwargs and kwargs['dtype'] == np.float64:
        with pytest.raises(TypeError):
            iotext.openmicrostates(
                limits_file=limits_file,
                file_name=traj_file,
                **kwargs,
            )
    elif 'no_traj' in kwargs:
        with pytest.raises(iotext.FileError):
            iotext.openmicrostates(
                limits_file=limits_file,
                file_name=traj_file,
            )
    else:
        # check with limits file
        expected = [[1, 1, 1, 1, 1], [2, 2, 1, 2, 3], [2, 2, 3]]
        trajs = iotext.openmicrostates(
            limits_file=limits_file,
            file_name=traj_file,
            **kwargs,
        )
        for i, traj in enumerate(trajs):
            assert (traj == expected[i]).all()


@pytest.mark.parametrize('traj_file, expected', [
    (
        os.path.join(HERE, 'traj1.dat'),
        [[1, 1, 1, 1, 1], [2, 2, 1, 2, 3], [2, 2, 3]],
    ),
    (
        os.path.join(HERE, 'data.dat'),
        [
            [[1, 2], [1, 3], [1, 2], [1, 1], [1, 2]],
            [[2, 1], [2, 5], [1, 4], [2, 3], [3, 2]],
            [[2, 1], [2, 2], [3, 3]],
        ],
    ),
])
def test_opentxt_limits(limits_file, traj_file, expected):
    """Test that the trajectory is split correctly."""
    # check with limits file
    trajs = iotext.opentxt_limits(
        limits_file=limits_file,
        file_name=traj_file,
    )
    for i, traj in enumerate(trajs):
        assert (traj == expected[i]).all()

    # check without limits file
    if len(trajs[0].shape) == 1:
        expected = [[1, 1, 1, 1, 1, 2, 2, 1, 2, 3, 2, 2, 3]]
        trajs = iotext.opentxt_limits(file_name=traj_file)
        for i, traj in enumerate(trajs):
            assert (traj == expected[i]).all()



# -*- coding: utf-8 -*-
"""Tests for the compare module.

BSD 3-Clause License
Copyright (c) 2019-2020, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest

from msmhelper import compare
from msmhelper.statetraj import StateTraj


@pytest.mark.parametrize('arr1, arr2, result', [
    ([0, 1, 2, 3], [4, 5, 6, 7], 0),
    ([0, 1, 2, 3], [1, 2, 3, 4], 3),
    ([0, 1, 2, 3], [0, 1, 2, 3], 4),
])
def test__intersect(arr1, arr2, result):
    """Test intersect method."""
    np.testing.assert_almost_equal(
        compare._intersect(arr1, arr2),
        result,
    )


@pytest.mark.parametrize('arr1, arr2, result', [(
    [np.array([0, 1, 2, 3, 4]), np.array([5, 6, 7])],
    [np.array([0, 1]), np.array([2, 3]), np.array([4, 5, 6, 7])],
    np.array([[2., 2., 1.], [0., 0., 3.]]),
)])
def test__intersect_array(arr1, arr2, result):
    """Test intersect method."""
    np.testing.assert_array_almost_equal(
        compare._intersect_array(arr1, arr2), result,
    )


@pytest.mark.parametrize('traj1, traj2, kwargs, result', [
    (
        [0, 0, 0, 0, 0, 1, 1, 1], [0, 0, 1, 1, 1, 2, 2, 2],
        {'method': 'directed'}, 1.0,
    ),
    (
        [0, 0, 1, 1, 1, 2, 2, 2], [0, 0, 0, 0, 0, 1, 1, 1],
        {'method': 'directed'}, 0.7,
    ),
    (
        [0, 0, 0, 0, 0, 1, 1, 1], [0, 0, 1, 1, 2, 2, 2, 2],
        {'method': 'symmetric'}, 0.90625,
    ),
    (
        [0, 0, 0, 0, 0, 1, 1, 1], [0, 0, 1, 1, 2, 2, 2, 2],
        {'method': 'directed'}, 0.8125,
    ),
    (
        [0, 0, 1, 1, 2, 2, 2, 2], [0, 0, 0, 0, 0, 1, 1, 1],
        {'method': 'directed'}, 0.6,
    ),
])
def test__compare_discretization(traj1, traj2, kwargs, result):
    """Test compare diescretitzation."""
    traj1, traj2 = StateTraj(traj1), StateTraj(traj2)
    np.testing.assert_almost_equal(
        compare._compare_discretization(traj1, traj2, **kwargs),
        result,
    )

    with pytest.raises(ValueError):
        compare._compare_discretization(traj1, traj2, method='NotAMethod')


@pytest.mark.parametrize('traj1, traj2, kwargs, error', [
    ([0, 0, 0, 0, 1], [0, 1, 1, 1, 1], {}, None),
    ([0, 0, 0, 0, 1], [0, 1, 1, 1, 1, 1], {}, ValueError),
    ([0, 0, 0, 0, 1], [0, 1, 1, 1], {}, ValueError),
    ([0, 0, 0, 0], [0, 1, 1, 1], {}, ValueError),
    ([0, 0, 0, 0, 1], [1, 1, 1], {}, ValueError),
    ([0, 0, 0, 0], [1, 1, 1], {}, ValueError),
    ([0, 0, 0, 0, 1], [0, 1, 1, 1, 1], {'method': 'symmetric'}, None),
    ([0, 0, 0, 0, 1], [0, 1, 1, 1, 1], {'method': 'directed'}, None),
    ([0, 0, 0, 0, 1], [0, 1, 1, 1, 1], {'method': 'forward'}, ValueError),
])
def test_compare_discretization(traj1, traj2, kwargs, error):
    """Test compare diescretitzation."""
    if error is None:
        compare.compare_discretization(traj1, traj2, **kwargs)
    else:
        with pytest.raises(error):
            compare.compare_discretization(traj1, traj2, **kwargs)
